/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Bootstrap code (restartless add-on)
 *
 * Note:
 * Toolbar button add code, partially taken from Dmitry Gutov's addon "Replace Bookmark"
 * see: https://addons.mozilla.org/fr/firefox/addon/bmreplace/
 *
 * ************************************************************************
 */

"use strict";

Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/AddonManager.jsm");
try {
  Components.utils.import("resource://gre/modules/devtools/Console.jsm");
} catch (ex) {
  // Older version of Mozilla (Firefox 13 for example), do not have the browser console.
  // Components.utils.reportError("[CookieKeeper] Error Console not available.");
}

let addonData  = null;
var currReason = null;
var self       = this;

function loadLib(data) {
  function include(aPath) {
    var path = addonData.resourceURI.spec + "lib/" + aPath + ".js";
    Services.scriptloader.loadSubScript(path, self);
  }

  if (addonData == null) {
    addonData = data;
    Services.scriptloader.loadSubScript(data.resourceURI.spec + "chrome/content/common.js", self);

    include("l10n");
    include("ui");
    include("preferences");
    include("storage");
    include("main");
  } else {
    addonData = data;
  }
}

function reportError(msg) {
  if (addonData != null) {
    Components.utils.reportError("[CookieKeeper" + addonData.version + "] " + msg);
  } else {
    Components.utils.reportError("[CookieKeeper] " + msg);
  }
}

function reasonToString(reason) {
  switch (reason) {
  case APP_STARTUP:
    return "APP_STARTUP";
  case ADDON_ENABLE:
    return "ADDON_ENABLE";
  case ADDON_INSTALL:
    return "ADDON_INSTALL";
  case ADDON_UPGRADE:
    return "ADDON_UPGRADE";
  case ADDON_DOWNGRADE:
    return "ADDON_DOWNGRADE";
  case APP_SHUTDOWN:
    return "APP_SHUTDOWN";
  case ADDON_DISABLE:
    return "ADDON_DISABLE";
  case ADDON_UNINSTALL:
    return "ADDON_UNINSTALL";
  }
  return "UNKNOWN (" + reason + ")";
}

function startup(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  APP_STARTUP
  /// &#10;  ADDON_ENABLE
  /// &#10;  ADDON_INSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  currReason = reason;
  try {
    loadLib(data);
    log.d("startup: " + reasonToString(reason));
    ckMain.init();
  } catch (ex) {
    reportError(ex);
  }
}

function shutdown(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  APP_SHUTDOWN
  /// &#10;  ADDON_DISABLE
  /// &#10;  ADDON_UNINSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  addonData  = data;
  currReason = reason;
  try {
    log.d("shutdown: " + reasonToString(reason));
    ckMain.uninit();

    if (reason != APP_SHUTDOWN) {
      // http://maglione-k.users.sourceforge.net/bootstrapped.xhtml
      let stringBundleService = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService);
      stringBundleService.flushBundles();
    }
  } catch (ex) {
    reportError(ex);
  }
}

function install(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  ADDON_INSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  currReason = reason;
  try {
    if (reason == ADDON_UPGRADE) {
      // Previous version of CookieKeeper do not flushed the string bundle cache.
      let Cc = Components.classes;
      let Ci = Components.interfaces;
      let stringBundleService = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService);
      stringBundleService.flushBundles();
    }

    loadLib(data);

    log.d("install: " + reasonToString(reason));
    ckPreferences.migrateCCPrefs();
    ckPreferences.updatePreferences();
  } catch (ex) {
    reportError(ex);
  }
}

function uninstall(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  ADDON_UNINSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  // Delete preferences
  if (ADDON_UNINSTALL == reason) {
    // Can't use the common part here, not available.
    Services.prefs.deleteBranch("extensions.cookiekeeper.");
  }
}
