/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Main window functions
 *
 * ************************************************************************
 */

"use strict";

const VIEW_TREE   = 0;
const VIEW_FLAT   = 1;
const VIEW_FILTER = 3;
// Protection filter values (show only all/protected/unprotected cookies):
const PROTECT_FILTER_ALL         = 0;
const PROTECT_FILTER_PROTECTED   = 1;
const PROTECT_FILTER_UNPROTECTED = 2;

const BUNDLE_PF_ALL           = "cookiesAll";
const BUNDLE_PF_ALL_PROT      = "cookiesAllProtected";
const BUNDLE_PF_ALL_UNPROT    = "cookiesAllUnprotected";
const BUNDLE_PF_FILTER        = "cookiesFiltered";
const BUNDLE_PF_FILTER_PROT   = "cookiesFilteredProtected";
const BUNDLE_PF_FILTER_UNPROT = "cookiesFilteredUnprotected";

const COL_DOMAIN     = "domainCol";
const COL_NAME       = "nameCol";
const COL_PATH       = "pathCol";
const COL_VALUE      = "valueCol";
const COL_PROTECTED  = "protectedCol";
const COL_HTTPONLY   = "httpCol";
const COL_SECURE     = "secureCol";
const COL_SESSION    = "sessionCol";
const COL_EXPIRE     = "expiresCol";
const COL_CREATION   = "creationCol";
const COL_LASTACCESS = "lastAccessedCol";

// Mouse button number
// const LEFT_BUTTON   = 0;
// const MIDDLE_BUTTON = 1;
// const RIGHT_BUTTON  = 2;

Cu.import("resource://cookiekeeper/cookiekeeper.jsm");

/**
 * Cookie Keeper Window management:
 */
var gCookieKeeper = {
  // Services
  _sCookieManager       : Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2),
  _sPreferences         : Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch),
  _sDateFormat          : Cc["@mozilla.org/intl/scriptabledateformat;1"].getService(Ci.nsIScriptableDateFormat),
  _sObserver            : Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService),
  // TODO: when it is possible to access dom storage
  // See bugs:
  // - https://bugzilla.mozilla.org/show_bug.cgi?id=343163
  // - https://bugzilla.mozilla.org/show_bug.cgi?id=506692
  // - https://bugzilla.mozilla.org/show_bug.cgi?id=527667
  //_sDomStorage          : Cc["@mozilla.org/dom/localStorage-manager;1"].getService(Ci.nsIDOMStorageManager),

  // cookies list
  _hosts                : {},
  _hostOrder            : [],

  // UI Variables
  _bundle               : null, // String translations (cookie culler specific)
  _tree                 : null, // Cookie XUL tree (cookiesTree)

  // String
  _yes                  : "yes",
  _no                   : "no",

  _labelProtec          : null, // Used to quickly set the right protect label for the context menu
  _labelRemove          : null, // Used to quickly set the right remove label for the context menu

  _blockCookieChangedEvent : false,
  _blockProtectionEvent    : false,

  init: function() {
    this._sObserver.addObserver(this, "cookie-changed", false);
    this._sObserver.addObserver(this, CK_OBS_PCOOKIE_TOPIC, false);

    this._bundle   = document.getElementById("bundleCookieKeeper");
    this._tree     = document.getElementById("cookiesTree");
    this._yes      = this._bundle.getString("yes");
    this._no       = this._bundle.getString("no");

    this._tree.treeBoxObject.view = this._view;

    gTestMozPrefs.doTest();

    this._refreshList(true);

    document.getElementById("textBoxFilter").focus();
  },

  uninit: function() {
    this._sObserver.removeObserver(this, "cookie-changed");
    this._sObserver.removeObserver(this, CK_OBS_PCOOKIE_TOPIC);
  },

  _makeStrippedHost: function (aHost) {
    var formattedHost = aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost;
    return formattedHost.substring(0, 4) == "www." ? formattedHost.substring(4, formattedHost.length) : formattedHost;
  },

  _addCookie: function (aStrippedHost, aCookie, aProtected) {
    // Check if the host exists, and it is not null
    if (!(aStrippedHost in this._hosts)) {
      this._hosts[aStrippedHost] = { cookies   : [],
                                     rawHost   : aStrippedHost,
                                     level     : 0,
                                     open      : false,
                                     container : true };
      this._hostOrder.push(aStrippedHost);
    }

    var cookie = this._makeCookieObject(aStrippedHost, aCookie, aProtected);
    this._hosts[aStrippedHost].cookies.push(cookie);

    return cookie;
  },

  _makeCookieObject: function (aStrippedHost, aCookie, aProtected) {
    // Some expiry value have I think the int64 max val: 0x7FFFFFFFFFFFFFFF
    // If so, set it to DATE_MAX instead

    var c = { name        : aCookie.name,
              value       : aCookie.value,
              isDomain    : aCookie.isDomain,
              host        : aCookie.host,
              rawHost     : aStrippedHost,
              path        : aCookie.path,
              isSecure    : aCookie.isSecure,
              expires     : aCookie.expires*1000,
              creationTime: aCookie.creationTime/1000,
              expiry      : aCookie.expiry > DATE_MAX ? DATE_MAX : aCookie.expiry*1000,
              isHttpOnly  : aCookie.isHttpOnly,
              isSession   : aCookie.isSession,
              lastAccessed: aCookie.lastAccessed/1000,
              isProtected : aProtected,
              level       : 1,
              container   : false };
    return c;
  },

  _loadCookies: function () {
    var e = this._sCookieManager.enumerator;
    var protCookies = CookieKeeper.loadProtectedCookies(this._sPreferences);
    this._hosts = {};
    this._hostOrder = [];

    // Load cookies
    while (e.hasMoreElements()) {
      var cookie = e.getNext();
      if (cookie && cookie instanceof Ci.nsICookie2) {
        // Search if it is a protected cookie
        var isProtected = false;
        for (let i=0; i<protCookies.length && !isProtected; i++) {
          var pc = protCookies[i];
          isProtected = pc.host == cookie.host && pc.name == cookie.name && pc.path == cookie.path;
        }

        // CK-1.6.0: Protection menu filter:
        if (!((isProtected && this._view._gFilter == PROTECT_FILTER_UNPROTECTED) ||
            (!isProtected && this._view._gFilter == PROTECT_FILTER_PROTECTED))) {
          var strippedHost = this._makeStrippedHost(cookie.host);
          this._addCookie(strippedHost, cookie, isProtected);
        }

      } else if (cookie) {
        log.e("non cookie class: " + cookie.constructor.name);
      } else {
        log.e("non cookie is null");
      }
    }
  },

  _formatDateString: function(aDate, defaultBundle) {
    if (aDate) {
      try {
        var date = new Date(aDate);
        return this._sDateFormat.FormatDateTime("", this._sDateFormat.dateFormatLong,
                                       this._sDateFormat.timeFormatSeconds,
                                       date.getFullYear(),
                                       date.getMonth() + 1,
                                       date.getDate(),
                                       date.getHours(),
                                       date.getMinutes(),
                                       date.getSeconds());
      } catch (e) {
        log.e("_formatDateString(" + aDate + ") exception:" + e);
        return "" + aDate;
      }
    } else if (typeof(defaultBundle) != "undefined" && defaultBundle != null) {
      return this._bundle.getString(defaultBundle);
    } else {
      return "<error " + aDate + ">";
    }
  },

  _updateCookieData: function () {
    var seln        = this._view.selection;
    var ids         = ["name", "value", "host", "path", "isSecure", "expires", "isProtected"];
    var properties  = null;
    var item        = null;
    var disabled    = true;

    if (this._view.mode == VIEW_TREE) {
      item = this._view._getItemAtIndex(seln.currentIndex);
    } else {
      item = this._view._flatSet[seln.currentIndex];
    }

    if (item && !item.container && seln.count > 0) {
      properties = { name: item.name, value: item.value, host: item.host,
                     path: item.path,
                     expires: item.isSession       ? this._bundle.getString("expireAtEndOfSession")
                                                   : this._formatDateString(item.expires, "expireAtEndOfSession"),
                     isDomain: item.isDomain       ? this._bundle.getString("domainColon")
                                                   : this._bundle.getString("hostColon"),
                     isSecure: item.isSecure       ? this._bundle.getString("forSecureOnly")
                                                   : this._bundle.getString("forAnyConnection"),
                     isProtected: item.isProtected ? this._yes
                                                   : this._no
                    };
      disabled = false;

    } else {
      var noneSelected = this._bundle.getString("noCookieSelected");
      properties = { name: noneSelected, value: noneSelected, host: noneSelected,
                     path: noneSelected, expires: noneSelected,
                     isSecure: noneSelected, isProtected: noneSelected };
    }

    for (let property in properties) {
      var elem      = document.getElementById(property);
      elem.value    = properties[property];
      elem.disabled = disabled;
    }
  },

  _updateButtonsProtect: function(aCookies, btProtec, btUnProt) {

    // Check if we have to protect or unprotect
    // if mixed content, choose protection
    const UNDEF_PROTECT = 0;
    const ALL_PROTEC    = 1;
    const ALL_UNPROTECT = 2;
    const MIXED_PROTEC  = 3;
    var   protection    = UNDEF_PROTECT;

    for (let i=0; i<aCookies.length && protection != MIXED_PROTEC; i++) {
      var cookie = aCookies[i];

      switch (protection) {
      case UNDEF_PROTECT:
        protection = (cookie.isProtected ? ALL_PROTEC : ALL_UNPROTECT);
        break;
      case ALL_PROTEC:
        protection = (cookie.isProtected ? ALL_PROTEC : MIXED_PROTEC);
        break;
      case ALL_UNPROTECT:
        protection = (cookie.isProtected ? MIXED_PROTEC : ALL_UNPROTECT);
        break;
      }
    }

    switch (protection) {
      case ALL_PROTEC:
        // all cookies protected, we can unprotect them at once
        btProtec.disabled = true;
        btUnProt.disabled = false;
        btUnProt.parentNode.selectedPanel = btUnProt;
        this._labelProtec = btUnProt.label;
        break;
      case ALL_UNPROTECT:
      case MIXED_PROTEC:
        btProtec.disabled = false;
        btUnProt.disabled = true;
        btProtec.parentNode.selectedPanel = btProtec;
        this._labelProtec = btProtec.label;
        break;
    }
  },

  _updateButtons: function () {
    var disabled = false;

    var tbFilter = document.getElementById("textBoxFilter"); // filter text box
    var btRemAll = document.getElementById("removeAllCookies");
    var btRemove = document.getElementById("removeCookie");
    var btRemMul = document.getElementById("removeCookies");
    var btProtec = document.getElementById("protectCookie");
    var btUnProt = document.getElementById("unprotectCookie");
    var btEdit   = document.getElementById("editButton");

    var seln = this._view.selection;

    if (this._view.mode == VIEW_FILTER) {
      disabled = this._view.rowCount == 0;
      tbFilter.disabled = false;
    } else {
      disabled = (Object.keys(this._hosts).length == 0);
      tbFilter.disabled = disabled;
      // debug:
      if (Object.keys(this._hosts).length > this._view.rowCount) {
        log.e("_updateButtons: row count error");
      }
    }

    btRemAll.disabled = disabled;
    btEdit.disabled   = true;

    if (disabled || seln.currentIndex == -1) {
      // No cookie is displayed
      // or no item is selected
      // disable all
      btRemove.disabled = true;
      btRemMul.disabled = true;
      btProtec.disabled = true;
      btUnProt.disabled = true;

      btRemove.parentNode.selectedPanel = btRemove;
      btProtec.parentNode.selectedPanel = btProtec;
      this._labelRemove = btRemove.label;
      this._labelProtec = btProtec.label;

    } else {
      // There is some cookies displayed, and at least one selected
      if (this._view.mode == VIEW_TREE) {
        // In tree view there is only one selected item max
        var item = this._view._getItemAtIndex(seln.currentIndex);
        // item can't be null here, or there is a bug somewhere
        if (item.container) {
          if (item.cookies == null || item.cookies.length == 0) {
            // debug
            log.e("_updateButtons: no cookie in container!");
          }

          if (item.cookies.length > 1) {
            btRemove.disabled = true;
            btRemMul.disabled = false;
            btRemMul.parentNode.selectedPanel = btRemMul;
            this._updateButtonsProtect(item.cookies, btProtec, btUnProt);
            this._labelRemove = btRemMul.label;
          } else {
            // One cookie in container
            btRemove.disabled = false;
            btRemMul.disabled = true;
            btRemove.parentNode.selectedPanel = btRemove;
            this._labelRemove = btRemove.label;
            if (item.cookies[0].isProtected) {
              btProtec.disabled = true;
              btUnProt.disabled = false;
              btUnProt.parentNode.selectedPanel = btUnProt;
              this._labelProtec = btUnProt.label;
            } else {
              btProtec.disabled = false;
              btUnProt.disabled = true;
              btProtec.parentNode.selectedPanel = btProtec;
              this._labelProtec = btProtec.label;
            }
          }

        } else {
          // item is a cookie
          btRemove.disabled = false;
          btRemMul.disabled = true;
          btEdit.disabled   = false;
          btRemove.parentNode.selectedPanel = btRemove;
          this._labelRemove = btRemove.label;
          if (item.isProtected) {
            btProtec.disabled = true;
            btUnProt.disabled = false;
            btUnProt.parentNode.selectedPanel = btUnProt;
            this._labelProtec = btUnProt.label;
          } else {
            btProtec.disabled = false;
            btUnProt.disabled = true;
            btProtec.parentNode.selectedPanel = btProtec;
            this._labelProtec = btProtec.label;
          }
        }

      } else {
        // flat & filter view
        var rangeCount          = seln.getRangeCount();
        var selectedCookieCount = 0;
        var cookies             = [];

        // Get all selected cookies
        for (let i = 0; i < rangeCount; i++) {
          var min = {}; var max = {};
          seln.getRangeAt(i, min, max);
          for (let j = min.value; j <= max.value; j++) {
            var cookie = this._view._getItemAtIndex(j);
            if (cookie == null) {
              log.e("cookie at index " + j + " is null");
              continue;
            }
            selectedCookieCount++;
            cookies.push(cookie);
          }
        }

        if (selectedCookieCount == 0) {
          log.e("_updateButtons: no cookie selected");
        } else if (selectedCookieCount == 1) {
          btRemove.disabled = false;
          btRemMul.disabled = true;
          btEdit.disabled   = false;
          btRemove.parentNode.selectedPanel = btRemove;
          this._labelRemove = btRemove.label;

          if (cookies[0].isProtected) {
            btProtec.disabled = true;
            btUnProt.disabled = false;
            btUnProt.parentNode.selectedPanel = btUnProt;
            this._labelProtec = btUnProt.label;
          } else {
            btProtec.disabled = false;
            btUnProt.disabled = true;
            btProtec.parentNode.selectedPanel = btProtec;
            this._labelProtec = btProtec.label;
          }
        } else {
          // more than 1 cookie selected
          btRemove.disabled = true;
          btRemMul.disabled = false;
          btRemMul.parentNode.selectedPanel = btRemMul;
          this._labelRemove = btRemMul.label;
          this._updateButtonsProtect(cookies, btProtec, btUnProt);
        }
      }
    }
  },

  _blockSelectEvent: 0,
  onCookieSelected: function () {
    if (this._blockSelectEvent == 0) {
      this._updateButtons();
      this._updateCookieData();
    } else if (this._blockSelectEvent < 0) {
      log.e("this._blockSelectEvent < 0: " + this._blockSelectEvent)
    }
  },

  performDeletion: function (deleteItems) {
    var blockFutureCookies = false;
    if (this._sPreferences.prefHasUserValue("network.cookie.blockFutureCookies")) {
      blockFutureCookies = this._sPreferences.getBoolPref("network.cookie.blockFutureCookies");
    }

    for (let i = 0; i < deleteItems.length; ++i) {
      var item = deleteItems[i];
      this._sCookieManager.remove(item.host, item.name, item.path, blockFutureCookies);
      if (item.isProtected) {
        CookieKeeper.removeProtectedCookie(item);
      }
    }
  },

  _askDelCookie: function(aCount) {

    var doNotAsk = false;
    var doDel    = true;

    if (aCount == 0) {
      return false;
    }

    if (this._sPreferences.prefHasUserValue(PREF_NO_ASK_AGAIN)) {
      doNotAsk = this._sPreferences.getBoolPref(PREF_NO_ASK_AGAIN);
    }

    if (!doNotAsk) {
      var prompts     = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
      var check       = {value: false};
      var flags       = prompts.BUTTON_TITLE_YES    * prompts.BUTTON_POS_0 +
                        prompts.BUTTON_TITLE_NO     * prompts.BUTTON_POS_1;
      var sDelProt    = this._bundle.getString(aCount > 1 ? "del_protects" : "del_protect");
      var sNoAskAgain = this._bundle.getString("no_ask_again");
      var button      = prompts.confirmEx(window, "CookieKeeper", sDelProt, flags,
                   "", "", "", sNoAskAgain, check);

      if (button == 0) {
        if (check.value) {
          // setup the preferences
          this._sPreferences.setBoolPref(PREF_NO_ASK_AGAIN, true);
        }
      } else {
        doDel = false;
      }
    }

    return doDel;
  },

  onDeleteCookie: function () {
    var seln = this._view.selection;
    if (seln.count < 1) {
      return;
    }

    var deleteItems = [];
    var nProtected   = 0;
    var delProtected = true;

    if (this._view.mode == VIEW_TREE) {
      var ci = seln.currentIndex;
      var item = this._view._getItemAtIndex(ci);

      if (typeof(item) == "undefined" || item == null) {
        log.e("No cookie at index " + ci);
        return;
      }

      if (item.container) {
        for (let i=0; i<item.cookies.length; i++) {
          var cookie = item.cookies[i];
          nProtected += cookie.isProtected ? 1 : 0;
          deleteItems.push(cookie);
        }
      } else {
        nProtected = item.isProtected ? 1 : 0;
        deleteItems.push(item);
      }
    } else {
      // VIEW_FLAT & VIEW_FILTER
      var rangeCount = seln.getRangeCount();

      for (let i = rangeCount - 1; i >= 0; i--) {
        var min = {}; var max = {};
        seln.getRangeAt(i, min, max);

        for (let j = min.value; j <= max.value; j++) {
          var item = this._view._getItemAtIndex(j);
          if (typeof(item) == "undefined" || item == null) {
            log.e("No cookie at index " + j);
            return;
          }

          nProtected += item.isProtected ? 1 : 0;
          deleteItems.push(item);
        }
      }
    }

    // common part:
    if (nProtected > 0) {
      delProtected = this._askDelCookie(nProtected);
    }

    this._blockSelectEvent++;
    if (delProtected) {
      this.performDeletion(deleteItems);
    } else {
      var deleteItems2 = [];
      for (let i=0; i<deleteItems.length; i++) {
        var cookie = deleteItems[i];
        if (!cookie.isProtected) {
          deleteItems2.push(cookie);
        }
      }
      this.performDeletion(deleteItems2);
    }
    this._blockSelectEvent--;
    this.onCookieSelected();
    this._tree.focus();
  },

  /// Delete all displayed and unprotected cookies
  onDeleteAllCookies: function () {
    const rowCount  = this._view.rowCount;
    const viewMode  = this._view.mode;
    var deleteItems = [];
    this._blockCookieChangedEvent = true;
    this._blockSelectEvent++;
    this._view.rowCount = 0;

    this._saveState();

    for (let key in this._hosts) {
      var host = this._hosts[key];
      const nc = host.cookies.length;
      for (let i=0; i<nc; i++) {
        var cookie = host.cookies[i];
        if (!cookie.isProtected &&
          (viewMode != VIEW_FILTER ||
            (viewMode == VIEW_FILTER && this._cookieMatchesFilter(cookie)))) {
              deleteItems.push(cookie);
         }
      }
    }

    this.performDeletion(deleteItems);

    this._refreshList(false);

    this._blockCookieChangedEvent = false;
    this._blockSelectEvent--;
    this.onCookieSelected();

    if (this._view.rowCount > 0) {
      this._tree.focus();
    } else {
      this.focusFilterBox();
    }
  },

  onProtectCookie: function(aEvent) {
    var seln = this._view.selection;
    if (seln.count == 0) {
      return;
    }
    const ci = seln.currentIndex;
    if (ci == -1) {
      return;
    }

    // use the button state, faster than check all cookies
    // to know if we have to protect or unprotect
    const doProtect = !document.getElementById("protectCookie").disabled;

    var tbo = this._tree.treeBoxObject;
    tbo.beginUpdateBatch();

    if (this._view.mode == VIEW_TREE) {
      var item = this._view._getItemAtIndex(ci);

      if (item.container) {
        if (!doProtect) {
          // Unprotect all cookies
          for (let i=0; i<item.cookies.length; i++) {
            var cookie = item.cookies[i];
            if (cookie.isProtected) {
              CookieKeeper.removeProtectedCookie(cookie);
              this._tree.treeBoxObject.invalidateRow(ci + i + 1);
            }
          }
        } else {
          // Protect all cookies
          for (let i=0; i<item.cookies.length; i++) {
            var cookie = item.cookies[i];
            if (!cookie.isProtected) {
              CookieKeeper.addProtectedCookie(cookie);
              this._tree.treeBoxObject.invalidateRow(ci + i + 1);
            }
          }
        }
      } else if (item.isProtected) {
        CookieKeeper.removeProtectedCookie(item);
        // Update list:
        this._tree.treeBoxObject.invalidateRow(ci);
      } else {
        CookieKeeper.addProtectedCookie(item);
        this._tree.treeBoxObject.invalidateRow(ci);
      }

    } else {
      // Flat & filter view
      var rangeCount = seln.getRangeCount();

      // For all selected cookies
      for (let i = 0; i < rangeCount; i++) {
        var min = {}; var max = {};
        seln.getRangeAt(i, min, max);
        for (let j = min.value; j <= max.value; j++) {
          var cookie = this._view._getItemAtIndex(j);
          if (cookie == null) {
            log.e("protectCookie: cookie at index " + j + " is null");
            continue;
          }

          if (doProtect && !cookie.isProtected) {
            CookieKeeper.addProtectedCookie(cookie);
            this._tree.treeBoxObject.invalidateRow(j);
          } else if (!doProtect && cookie.isProtected) {
            CookieKeeper.removeProtectedCookie(cookie);
            this._tree.treeBoxObject.invalidateRow(j);
          }
        }
      }

    }

    // Update ui:
    tbo.endUpdateBatch();
    this.onCookieSelected();
    this._tree.focus();
  },

  onShowEditor: function() {
    var seln = this._view.selection;
    if (seln.count == 0) {
      return;
    }
    const ci = seln.currentIndex;
    if (ci == -1) {
      return;
    }

    var item = this._view._getItemAtIndex(ci);
    if (item.container) {
      return;
    }

    window.openDialog("chrome://cookiekeeper/content/cookie.editor.xul",
      "dlgEditCookie",
      "chrome,centerscreen,resizable,modal",
      item
    );
    this._tree.focus();
  },

  onCookieKeyPress: function (aEvent) {
    switch (aEvent.charCode) {
      case aEvent.DOM_VK_SPACE:
        this.onProtectCookie(aEvent);
        aEvent.stopPropagation();
        return;
      default:
        break;
    }
    switch (aEvent.keyCode) {
      case aEvent.DOM_VK_DELETE:
        this.onDeleteCookie();
        aEvent.stopPropagation();
        return;
      default:
        break;
    }
  },

  onWindowKeyPress: function (aEvent) {
    if (aEvent.keyCode == KeyEvent.DOM_VK_ESCAPE) {
      window.close();
    }
  },

  onContextMenu: function (aEvent) {
    // Prevent to show the context menu if there is no selection, or nothing displayed
    // gCookiesTreeView.selection.count may return 1, even if the tree view is empty...
    var showContextMenu = (this._view.rowCount > 0 && this._view.selection.count > 0);
    if (!showContextMenu) {
      // Seems to work better to prevent the popup showing, than return false.
      aEvent.preventDefault();
    }
    return showContextMenu;
  },

  onShowContextCookieMenu: function(aEvent) {
    // The context menu can't show up if there is no selection
    // see onContextMenu() function
    var seln = this._view.selection;

    var mpccPerms      = document.getElementById("mpcc-perms");
    var mpccSepPerm    = document.getElementById("mpcc-sep-perm");

    var mpccProtect    = document.getElementById("mpcc-protect");
    var mpccEdit       = document.getElementById("mpcc-edit");
    var mpccRemove     = document.getElementById("mpcc-remove");

    var mpccSepCopy    = document.getElementById("mpcc-sep-copy");
    var mpccCopyDomain = document.getElementById("mpcc-copy-domain");
    var mpccCopyName   = document.getElementById("mpcc-copy-name");
    var mpccCopyValue  = document.getElementById("mpcc-copy-value");

    // Update the context menu label/hidden properties:
    mpccProtect.label = this._labelProtec;
    mpccRemove.label  = this._labelRemove;

    if (this._view.mode == VIEW_TREE) {
      var ci = seln.currentIndex;
      var item = this._view._getItemAtIndex(ci);

      mpccPerms.hidden   = false;
      mpccSepPerm.hidden = false;

      if (item.container) {
        mpccEdit.hidden       = true;
        mpccSepCopy.hidden    = false;
        mpccCopyDomain.hidden = false;
        mpccCopyName.hidden   = true;
        mpccCopyValue.hidden  = true;

      } else { // a cookie
        mpccEdit.hidden       = false;
        mpccSepCopy.hidden    = false;
        mpccCopyDomain.hidden = false;
        mpccCopyName.hidden   = false;
        mpccCopyValue.hidden  = false;
      }

    } else {
      // Flat view
      if (seln.count > 1) {
        mpccPerms.hidden   = true;
        mpccSepPerm.hidden = true;

        mpccEdit.hidden       = true;
        mpccSepCopy.hidden    = true;
        mpccCopyDomain.hidden = true;
        mpccCopyName.hidden   = true;
        mpccCopyValue.hidden  = true;
      } else {
        mpccPerms.hidden   = false;
        mpccSepPerm.hidden = false;

        mpccEdit.hidden       = false;
        mpccSepCopy.hidden    = false;
        mpccCopyDomain.hidden = false;
        mpccCopyName.hidden   = false;
        mpccCopyValue.hidden  = false;
      }
    }

    return true;
  },

  onShowContextCookiePerms: function(aEvent) {
    var sIO    = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    var sPerm  = Cc["@mozilla.org/permissionmanager;1"].getService(Ci.nsIPermissionManager);
    var sCPerm = Cc["@mozilla.org/cookie/permission;1"].getService(Ci.nsICookiePermission);

    var seln = this._view.selection;
    var ci   = seln.currentIndex;
    var item = this._view._getItemAtIndex(ci);
    var aHost = item.rawHost;

    var mpccDefault = document.getElementById("mpcc-perm-default");
    var mpccAllow   = document.getElementById("mpcc-perm-allow");
    var mpccDeny    = document.getElementById("mpcc-perm-deny");
    var mpccSession = document.getElementById("mpcc-perm-session");
    // var mpccFirst   = document.getElementById("mpcc-perm-first");
    // var mpccThird   = document.getElementById("mpcc-perm-third");

    mpccDefault.setAttribute("checked", false);
    mpccAllow.setAttribute("checked", false);
    mpccDeny.setAttribute("checked", false);
    mpccSession.setAttribute("checked", false);
    // mpccFirst.setAttribute("checked", false);
    // mpccThird.setAttribute("checked", false);

    if (!item.container) {
      aHost = item.host;
    }
    aHost = aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost;

    var uri  = sIO.newURI("http://" + aHost, null, null);
    var perm = sPerm.testPermission(uri, "cookie");

    switch (perm) {
    case sCPerm.ACCESS_DEFAULT:
      mpccDefault.setAttribute("checked", true);
      break;
    case sCPerm.ACCESS_ALLOW:
      mpccAllow.setAttribute("checked", true);
      break;
    case sCPerm.ACCESS_DENY:
      mpccDeny.setAttribute("checked", true);
      break;
    case sCPerm.ACCESS_SESSION:
      mpccSession.setAttribute("checked", true);
      break;
    // case sCPerm.ACCESS_ALLOW_FIRST_PARTY_ONLY:
    //   mpccFirst.setAttribute("checked", true);
    //   break;
    // case sCPerm.ACCESS_LIMIT_THIRD_PARTY:
    //   mpccThird.setAttribute("checked", true);
    //   break;
    }

    return true;
  },

  onCookieDBLClick: function (aEvent) {
    var seln = this._view.selection;
    if (seln.count == 0) {
      return;
    }
    var ci = seln.currentIndex;
    var item = this._view._getItemAtIndex(ci);
    if (!item.container) {
      this.onProtectCookie(aEvent);
    }
  },

  onContextMenuPermission: function(aPermission) {
    var sIO    = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    var sPerm  = Cc["@mozilla.org/permissionmanager;1"].getService(Ci.nsIPermissionManager);
    var sCPerm = Cc["@mozilla.org/cookie/permission;1"].getService(Ci.nsICookiePermission);

    var seln = this._view.selection;
    var ci   = seln.currentIndex;
    var item = this._view._getItemAtIndex(ci);
    var aHost = item.rawHost;

    if (!item.container) {
      aHost = item.host;
    }
    aHost = aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost;

    var uri = sIO.newURI("http://" + aHost, null, null);
    sPerm.add(uri, "cookie", aPermission);
  },

  onContextMenuCopy: function(anItem) {
    var sClipboard = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper);
    var seln = this._view.selection;
    var ci = seln.currentIndex;
    var item = this._view._getItemAtIndex(ci);

    if (anItem == "domain") {
      if (item.container) {
        sClipboard.copyString(item.rawHost);
      } else {
        sClipboard.copyString(item.host);
      }
    } else if (anItem == "name") {
      sClipboard.copyString(item.name);
    } else {
      sClipboard.copyString(item.value);
    }
  },

  _lastSortColumn   : COL_DOMAIN,
  _lastSortProperty : "host",
  _lastSortAscending: true,
  sort: function (aEvent, aProperty, aColumn) {
    if (aEvent != null && aEvent.button != 0) {
      return;
    }
    var ascending = (aProperty == this._lastSortProperty) ? !this._lastSortAscending : true;

    function sortString(a, b) {
      var c = a.toLowerCase().localeCompare(b.toLowerCase());
      if (ascending) { return c; }
      else { return -c; }
    }

    function sortCookieHost(a, b) {
      var c = a.host.toLowerCase().localeCompare(b.host.toLowerCase());
      if (c == 0) {
        c = a.name.toLowerCase().localeCompare(b.name.toLowerCase());
      }
      if (ascending) { return c; }
      else { return -c; }
    }

    function sortCookieName(a, b) {
      var c = a.name.toLowerCase().localeCompare(b.name.toLowerCase());
      if (c == 0) {
        c = a.host.toLowerCase().localeCompare(b.host.toLowerCase());
      }
      if (ascending) { return c; }
      else { return -c; }
    }

    function sortCookieString(a, b) {
      var c = a[aProperty].toLowerCase().localeCompare(b[aProperty].toLowerCase());
      if (c == 0) { return sortCookieHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    function sortCookieBool(a, b) {
      var c = a[aProperty] - b[aProperty];
      if (c == 0) { return sortCookieHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    function sortCookieDate(a, b) {
      var c = a[aProperty] - b[aProperty];
      if (c == 0) { return sortCookieHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    function sortCookieExpireDate(a, b) {
      var da = a.isSession ? a.expiry : a.expires;
      var db = b.isSession ? b.expiry : b.expires;
      var c = da - db;
      if (c == 0) { return sortCookieHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    var method = sortCookieBool;
    if (aProperty == "host") {
      method = sortCookieHost;
    } else if (aProperty == "name") {
      method = sortCookieName;
    } else if (aProperty == "path" || aProperty == "value") {
      method = sortCookieString;
    } else if (aProperty == "creationTime" || aProperty == "lastAccessed") {
      method = sortCookieDate;
    } else if (aProperty == "expires") {
      method = sortCookieExpireDate;
    }

    function buildFlatSet() {
      // Add all cookies
      var flatSet = [];
      var hosts   = gCookieKeeper._hosts;
      for (let key in hosts) {
        var host = hosts[key];
        flatSet = flatSet.concat(host.cookies);
      }
      flatSet.sort(method);

      // Update ui
      gCookieKeeper._view._flatSet = flatSet;
      gCookieKeeper._view.rowCount = flatSet.length;

      if (gCookieKeeper._view.rowCount != flatSet.length) {
        log.e("FlatSet: Error: have " + gCookieKeeper._view.rowCount + " expected " + flatSet.length);
      }
    }

    var tbo = this._tree.treeBoxObject;
    tbo.beginUpdateBatch();
    this._blockSelectEvent++;

    const viewMode = this._view.mode;

    if (viewMode != VIEW_FILTER && aProperty == "host") {
      // The sort is only by hosts
      this._hostOrder.sort(sortString);
      // Sort cookies names
      for (let key in this._hosts) {
        var host = this._hosts[key];
        host.cookies.sort(method);
      }

      // Change from FLAT to TREE?
      if (viewMode == VIEW_FLAT) {
        // Restore open cookies
        this._view.mode = VIEW_TREE;
        this._view.rowCount = Object.keys(this._hosts).length;
      }

    } else if (viewMode == VIEW_TREE) {
      // Change from TREE to FLAT
      this._saveState();
      this._view.mode = VIEW_FLAT;
      buildFlatSet();

    } else if (viewMode == VIEW_FLAT && this._view._flatSet.length == 0 && Object.keys(this._hosts).length > 0) {
        // Ensure to display cookies if we have some (call by _refreshList() may be in that case)
        // Like after clearing the filter
        //this._lastSortAscending = !this._lastSortAscending;
        buildFlatSet();
    } else if (this._view._flatSet.length > 0) {
      // mode == VIEW_FILTER
      this._view._flatSet.sort(method);
    }

    // Adjust the Sort Indicator
    var oldCol          = document.getElementById(this._lastSortColumn);
    var newCol          = document.getElementById(aColumn);
    var sortOrderString = ascending ? "ascending" : "descending";

    if (oldCol == null) {
      oldCol = document.getElementById(COL_DOMAIN);
      log.e("Old column unknown: " + aColumn);
    }
    if (newCol == null) {
      newCol = document.getElementById(COL_DOMAIN);;
      log.e("New column unknown: " + aColumn);
    }

    oldCol.removeAttribute("sortDirection");
    newCol.setAttribute("sortDirection", sortOrderString);

    this._view._invalidateCache(0);
    if (this._view.selection != null) {
      this._view.selection.clearSelection();
      this._view.selection.select(0);
    }

    this._tree.treeBoxObject.invalidate();
    this._tree.treeBoxObject.ensureRowIsVisible(0);

    this._lastSortAscending = ascending;
    this._lastSortProperty  = aProperty;
    this._lastSortColumn    = aColumn;

    this._restoreState();
    tbo.endUpdateBatch();

    this._blockSelectEvent--;
    this.onCookieSelected();
  },

  onSortByColNumber: function(aNumber) {
    var treeCols = document.getElementById("cookiesTreeCols").childNodes;
    var col      = null;
    var colNum   = 0;
    // Search for the first display column number:
    for (var i=0; i<treeCols.length; i++) {
      let obj = treeCols[i];
      if (obj.nodeName == "treecol" && obj.hidden == false) {
        if (colNum == aNumber) {
          col = obj;
          break;
        }
        colNum++;
      }
    }
    if (col != null) {
      this._tree.disabled = true;
      col.click();
      this._tree.disabled = false;
    }
  },

  _refreshList: function(aInitialLoad) {
    // Just reload the list to make sure deletions are respected
    this._loadCookies();
    // set the tree view: will assign _view.selection object.

    var labelPFilter = document.getElementById("labelProtectionFilter");

    if (aInitialLoad) {
      this._view.mode         = VIEW_TREE;
      this._lastSortAscending = false;

      this._view.rowCount = Object.keys(this._hosts).length;

      switch (this._view._gFilter) {
        case PROTECT_FILTER_ALL:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL);
          break;
        case PROTECT_FILTER_PROTECTED:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL_PROT);
          break;
        case PROTECT_FILTER_UNPROTECTED:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL_UNPROT);
          break;
      }

      this.sort(null, "host", COL_DOMAIN);
    } else if (this._view.mode == VIEW_FILTER) {
      this.onFilter();
    } else {
      this._view.mode = this._view.mode; // Be sure to reset the view (via get/set)
      this._view.rowCount = Object.keys(this._hosts).length;

      switch (this._view._gFilter) {
        case PROTECT_FILTER_ALL:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL);
          break;
        case PROTECT_FILTER_PROTECTED:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL_PROT);
          break;
        case PROTECT_FILTER_UNPROTECTED:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL_UNPROT);
          break;
      }

      this._lastSortAscending = !this._lastSortAscending;
      this.sort(null, this._lastSortProperty, this._lastSortColumn);
    }
  },

  onProtectionFilter: function(aFilterValue) {
    this._saveState();
    this._blockSelectEvent++;

    if (aFilterValue == "all") {
      this._view._gFilter = PROTECT_FILTER_ALL;
    } else if (aFilterValue == "unprotected") {
      this._view._gFilter = PROTECT_FILTER_UNPROTECTED;
    } else {
      this._view._gFilter = PROTECT_FILTER_PROTECTED;
    }

    this._refreshList(false);
    this._restoreState();
    this._blockSelectEvent--;
    // Update buttons
    this.onCookieSelected();
    this._tree.focus();
  },

  clearFilter: function () {
    if (this._view.mode == VIEW_FILTER) {
      this._blockSelectEvent++;
      this._view.mode = this._lastSortProperty == "host" ? VIEW_TREE : VIEW_FLAT;
      this._refreshList(false);
      this._blockSelectEvent--;
    }
  },

  _cookieMatchesFilter: function (aCookie) {
    return aCookie.rawHost.indexOf(this._view._filterValue) != -1 ||
           aCookie.name.indexOf(this._view._filterValue) != -1 ||
           aCookie.value.indexOf(this._view._filterValue) != -1;
  },

  _filterCookies: function (aFilterValue) {
    var cookies = [];
    for (let i = 0; i < this._hostOrder.length; ++i) {
      var currHost = this._hosts[this._hostOrder[i]];
      for (let j = 0; j < currHost.cookies.length; ++j) {
        var cookie = currHost.cookies[j];
        if (this._cookieMatchesFilter(cookie))
          cookies.push(cookie);
      }
    }
    return cookies;
  },

  onFilter: function () {
    var filter = document.getElementById("textBoxFilter").value;
    if (filter.length == 0) {
      this.clearFilter();
      return;
    }

    this._saveState();
    this._view.mode         = VIEW_FILTER;
    this._view._filterValue = filter;
    this._view._flatSet     = this._filterCookies(filter);

    // Move to multi-select in the tree
    this._tree.setAttribute("seltype", "multiple");

    // Clear & Set up the filtered display
    this._view.rowCount = this._view._flatSet.length;

    // if the view is not empty then select the first item
    if (this._view.rowCount > 0) {
      // sort the filtered cookies
      this._lastSortAscending = !this._lastSortAscending;
      this.sort(null, this._lastSortProperty, this._lastSortColumn);
      this._view.selection.select(0);
    }

    var labelPFilter = document.getElementById("labelProtectionFilter");
    switch (this._view._gFilter) {
      case PROTECT_FILTER_ALL:
        labelPFilter.value = this._bundle.getString(BUNDLE_PF_FILTER);
        break;
      case PROTECT_FILTER_PROTECTED:
        labelPFilter.value = this._bundle.getString(BUNDLE_PF_FILTER_PROT);
        break;
      case PROTECT_FILTER_UNPROTECTED:
        labelPFilter.value = this._bundle.getString(BUNDLE_PF_FILTER_UNPROT);
        break;
    }
    this._updateButtons();
  },

  focusFilterBox: function () {
    var filter = document.getElementById("textBoxFilter");
    filter.focus();
    filter.select();
    filter.disabled = (Object.keys(this._hosts).length <= 1);
  },

  _selectedItem: null,
  _openHosts: [],
  _saveState: function () {
    // Save selection
    if (this._view.rowCount > 0 && this._view.selection.currentIndex != -1) {
      this._selectedItem = this._view._getItemAtIndex(this._view.selection.currentIndex);
    }
    // Save open states
    if (this._view.mode == VIEW_TREE) {
      this._openHosts = [];
      for (let key in this._hosts) {
        var host = this._hosts[key];
        if (host.open) {
          this._openHosts.push(host.rawHost);
        }
      }
    }
  },

  _restoreState: function() {
    if (this._view.mode == VIEW_TREE && this._openHosts.length > 0) {
      for (let i=0; i<this._view.rowCount; i++) {
        var item = this._view._getItemAtIndex(i);
        if (item.container) {
          item.open = false;
          if (this._openHosts.indexOf(item.rawHost) > -1) {
            this._view.toggleOpenState(i);
          }
        }
      }
      this._openHosts = [];
    }

    if (this._selectedItem != null) {
      let idx = this._view._getIndexOfItem(this._selectedItem);
      if (idx != -1) {
        this._view.selection.select(idx);
        this._tree.treeBoxObject.ensureRowIsVisible(idx);
        this._selectedItem = null;
      }
    }
  }

} // var gCookieKeeper
