/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Observer:
 *  - Cookie add/remove/change.
 *  - Protected cookie add/remove.
 *
 * ************************************************************************
 */

"use strict";

gCookieKeeper.observe = function(aCookie, aTopic, aData) {

  if (!gCookieKeeper._blockProtectionEvent && aTopic == CK_OBS_PCOOKIE_TOPIC) {
    // somewhat a strange object: wrappedJSObject ...
    if (aData == CK_OBS_PCOOKIE_DEL) {
      this.onProtCookieAddRemoved(aCookie.wrappedJSObject.pcookie, false);
    } else if (aData == CK_OBS_PCOOKIE_ADD) {
      this.onProtCookieAddRemoved(aCookie.wrappedJSObject.pcookie, true);
    }
    return;
  }

  if (gCookieKeeper._blockCookieChangedEvent || aTopic != "cookie-changed") {
    return;
  }

  if (aCookie instanceof Ci.nsICookie2) {
    var strippedHost = this._makeStrippedHost(aCookie.host);
    if (aData == "changed") {
      this._handleCookieChanged(aCookie, strippedHost);
    } else if (aData == "added") {
      this._handleCookieAdded(aCookie, strippedHost);
    } else if (aData == "deleted") {
      this._handleCookieDeleted(aCookie, strippedHost);
    }
  } else if (aData == "cleared") {
    this._hosts = {};
    this._hostOrder = [];

    this._view.rowCount = 0;
    this._view.selection.clearSelection();
    this._updateButtons();
  } else if (aData == "reload") {
    // first, clear any existing entries
    this.observe(aCookie, aTopic, "cleared");

    // then, reload the list
    this._populateList(false);
  }
}

gCookieKeeper._handleCookieChanged = function(changedCookie, strippedHost) {
  function cookieEquals(a , b) {
    return a.host === b.host &&
           a.name === b.name &&
           a.path === b.path;
  }

  function updateCookie(aSource, aTarget) {
    aTarget.value        = aSource.value;
    aTarget.path         = aSource.path;
    aTarget.isSecure     = aSource.isSecure;
    aTarget.isDomain     = aSource.isDomain;
    aTarget.isSession    = aSource.isSession;
    aTarget.isHttpOnly   = aSource.isHttpOnly;
    aTarget.expires      = aSource.expires*1000;
    aTarget.expiry       = aSource.expiry > DATE_MAX ? DATE_MAX : changedCookie.expiry*1000;
    aTarget.creationTime = aSource.creationTime/1000;
    aTarget.lastAccessed = aSource.lastAccessed/1000;
  }

  var rowIndex = -1;
  var cookieItem = null;
  if (this._view.mode == VIEW_TREE) {
    for (let i = 0; i < this._hostOrder.length && cookieItem==null; i++) {
      rowIndex++;
      var hostItem = this._hosts[this._hostOrder[i]];
      if (this._hostOrder[i] == strippedHost) {
        // Host matches, look for the cookie within this Host collection
        // and update its data
        for (let j = 0; j < hostItem.cookies.length; j++) {
          if (hostItem.open) {
            rowIndex++;
          }
          var currCookie = hostItem.cookies[j];
          if (cookieEquals(currCookie, changedCookie)) {
            updateCookie(changedCookie, currCookie);
            cookieItem = currCookie;
            break;
          }
        }
      } else if (hostItem.open) {
        rowIndex += hostItem.cookies.length;
      }
    }
  } else {
    // Just walk the filter list to find the item. It doesn't matter that
    // we don't update the main Host collection when we do this, because
    // when the filter is reset the Host collection is rebuilt anyway.
    for (rowIndex = 0; rowIndex < this._view._flatSet.length; rowIndex++) {
      currCookie = this._view._flatSet[rowIndex];
      if (cookieEquals(currCookie, changedCookie)) {
        updateCookie(changedCookie, currCookie);
        cookieItem = currCookie;
        break;
      }
    }
  }

  // Make sure the tree display is up to date...
  this._tree.treeBoxObject.invalidateRow(rowIndex);
  // ... and if the cookie is selected, update the displayed metadata too
  if (cookieItem != null && this._view.selection.currentIndex == rowIndex) {
    this._updateCookieData();
  }
}

gCookieKeeper._handleCookieAdded = function(addedCookie, strippedHost) {
  var rowCountImpact = 0;
  // TODO: Optimize: Don't load every times the prot cookie list
  var isProtected = CookieKeeper.isProtectedCookie(addedCookie, null);
  var cookie = this._addCookie(strippedHost, addedCookie, isProtected);
  if (this._view.mode == VIEW_TREE) {
    // The Host collection for this cookie already exists, and it's not open,
    // so don't increment the rowCountImpact becaues the user is not going to
    // see the additional rows as they're hidden.
    let host = this._hosts[strippedHost];
    if (host.open || host.cookies.length == 1) {
      rowCountImpact++;
    }
  } else if (this._view.mode == VIEW_FLAT || this._cookieMatchesFilter(cookie)) {
    rowCountImpact++;
    this._view._flatSet.push(cookie);
  }

  if (rowCountImpact > 0) {
    // Now update the tree display at the end
    var oldRowCount = this._rowCount;
    var tbo = this._tree.treeBoxObject;
    tbo.beginUpdateBatch();

    this._view._rowCount += rowCountImpact;
    tbo.rowCountChanged(oldRowCount - 1, rowCountImpact);
    // Sort the list
    this._lastSortAscending = !this._lastSortAscending;
    this.sort(null, this._lastSortProperty, this._lastSortColumn);
    tbo.endUpdateBatch();
    // Update buttons status
    this._updateButtons();
  }
}

gCookieKeeper._handleCookieDeleted = function(aCookie, strippedHost) {
  // Search for the cookie:
  var currHost = this._hosts[strippedHost];
  if (typeof(currHost) == "undefined" || currHost == null) {
    log.e("_handleCookieDeleted: Cookie host " + strippedHost + " not found");
    return;
  }
  var cookie  = null;
  var cCookie = 0;
  const nCookies = currHost.cookies.length;
  for (; cCookie<nCookies; cCookie++) {
    var c = currHost.cookies[cCookie];
    if (c != null && c.host == aCookie.host && c.name == aCookie.name && c.path == aCookie.path) {
      cookie = c;
      break;
    }
  }

  if (cookie == null) {
    log.e("_handleCookieDeleted: Cookie not found: " + aCookie.host + ":" + aCookie.name);
    return;
  }
  const cookieViewIndex = this._view._getIndexOfItem(cookie);

  var idxHO = -1;
  if (nCookies == 1) {
    // Remote the host from _hostOrder
    idxHO = this._hostOrder.indexOf(strippedHost);
    if (idxHO > -1) {
      this._hostOrder.splice(idxHO, 1);
    } // else it is an error
    // Remove the host from the _hosts list
    delete this._hosts[strippedHost];
  } else {
    currHost.cookies.splice(cCookie, 1);
  }

  var tbo = this._tree.treeBoxObject;
  var seln = this._view.selection;
  const curSelIndex = seln.currentIndex;

  tbo.beginUpdateBatch();

  switch (this._view.mode) {
  case VIEW_TREE:
    if (cookieViewIndex == -1 && idxHO == -1) {
      // not displayed
      tbo.endUpdateBatch();
      return;
    }
    this._blockSelectEvent++;
    // TODO: a more fine update?
    var rowCountImpact = 0;
    var rowCount = this._view.rowCount;

    if (currHost.open) {
      rowCountImpact--;
    }
    if (idxHO > -1) {
      rowCountImpact--;
    }

    if (rowCountImpact != 0) {
      rowCount += rowCountImpact;
      this._view.rowCount = rowCount;
      // Update selection:
      seln.clearSelection();
      this._view._invalidateCache(0);
      tbo.invalidate();
      this._blockSelectEvent--;

      var newSelIndex = curSelIndex;
      if (cookieViewIndex != curSelIndex) {
        // The host is selected, not the cookie itself
      } else if (!currHost.open || (currHost.open && (nCookies == 1 || (nCookies > 1 && cCookie == (nCookies - 1))))) {
        // very complicated if ...
        newSelIndex--;
      }

      // Check limits:
      newSelIndex = (newSelIndex >= rowCount) ? rowCount - 1 : newSelIndex;
      newSelIndex = (newSelIndex < 0) ? 0 : newSelIndex;
      if (rowCount >= 0) {
        seln.select(newSelIndex);
        tbo.ensureRowIsVisible(newSelIndex);
      }
    } else {
      this._blockSelectEvent--;
    }
    break;
  case VIEW_FILTER:
    if (!this._cookieMatchesFilter(cookie)) {
      // not displayed
      break;
    }
    // pass through
  case VIEW_FLAT:
    this._blockSelectEvent++;

    this._view._flatSet.splice(cookieViewIndex, 1);
    this._view._rowCount--;
    tbo.invalidateRow(cookieViewIndex);
    tbo.rowCountChanged(cookieViewIndex, -1);

    // Update selection
    if (curSelIndex >= cookieViewIndex) {
      seln.clearSelection();
      this._blockSelectEvent--;
      if (curSelIndex >= this._view._flatSet.length) {
        seln.select(curSelIndex - 1);
      } else {
        seln.select(curSelIndex);
      }
    } else {
      this._blockSelectEvent--;
    }
    break;
  }

  tbo.endUpdateBatch();
}

gCookieKeeper.onProtCookieAddRemoved = function(pCookie, isAdded) {
  // Call from gCookieKeeperProtCookiesPrefs.remove()
  if (!pCookie) {
    return;
  }

  var strippedHost = this._makeStrippedHost(pCookie.host);
  var host = this._hosts[strippedHost];
  if (typeof(host) != "undefined") {
    const nCookies = host.cookies.length;
    for (let i=0; i<nCookies; i++) {
      var cookie = host.cookies[i];
      if (cookie.host == pCookie.host && cookie.name == pCookie.name && cookie.path == pCookie.path) {
        cookie.isProtected = isAdded;

        // Update ui:
        var index = this._view._getIndexOfItem(cookie);
        if (index > -1) {
          this._tree.treeBoxObject.invalidateRow(index);
          if (this._view.selection.currentIndex == index) {
            this.onCookieSelected();
          }
        }
        i = nCookies;
      }
    }
  }
}
