/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Cookies tree view.
 *
 * ************************************************************************
 */

"use strict";

gCookieKeeper._view = {
  _mode       : VIEW_TREE,  // _filtered (bool) in original code
  _flatSet    : [],         // _filterSet in original code
  _filterValue: "",
  _gFilter    : PROTECT_FILTER_ALL,  // protection filter
  _rowCount   : 0,
  _cacheValid : 0,
  _cacheItems : [],
  _selection: null, // setup automatically when this _view model is attach to the tree view

  get rowCount() {
    return this._rowCount;
  },

  set rowCount(val) {
    const oldCount = this._rowCount;
    if (val != oldCount) {
      this._rowCount = 0;
      gCookieKeeper._tree.treeBoxObject.rowCountChanged(0, -oldCount);
      if (val > 0) {
        this._rowCount = val;
        gCookieKeeper._tree.treeBoxObject.rowCountChanged(0, val);
      }
    }
  },

  get mode() {
    return this._mode;
  },

  set mode(val) {
    this._mode        = val;
    this.rowCount     = 0;
    this._cacheValid  = 0;
    this._flatSet     = [];
    this._cacheItems  = [];
    this._filterValue = "";

    if (val == VIEW_TREE) {
      gCookieKeeper._tree.setAttribute("seltype", "single");
    } else if ((val & VIEW_FLAT) == VIEW_FLAT) {
      gCookieKeeper._tree.setAttribute("seltype", "multiple");
    } else {
      log.e("_view: unknown view mode " + val);
    }
  },

  _getItemAtIndex: function (aIndex) {
    if ((this._mode & VIEW_FLAT) == VIEW_FLAT) {
      if (aIndex > -1 && aIndex < this._flatSet.length) {
        return this._flatSet[aIndex];
      } else {
        return null;
      }
    }

    var start = 0;
    var count = 0;
    var hostIndex = 0;

    var cacheIndex = Math.min(this._cacheValid, aIndex);
    if (cacheIndex > 0) {
      var cacheItem = this._cacheItems[cacheIndex];
      start = cacheItem['start'];
      count = hostIndex = cacheItem['count'];
    }

    const hoLength = gCookieKeeper._hostOrder.length;
    for (let i = start; i < hoLength; i++) {
      var currHost = gCookieKeeper._hosts[gCookieKeeper._hostOrder[i]];
      if (count == aIndex) { return currHost; }
      hostIndex = count;

      var cacheEntry = { 'start' : i, 'count' : count };
      var cacheStart = count;

      if (currHost.open) {
        if (count < aIndex && aIndex <= (count + currHost.cookies.length)) {
          // We are looking for an entry within this host's children,
          // enumerate them looking for the index.
          count++;
          for (let j = 0; j < currHost.cookies.length; j++) {
            if (count == aIndex) {
              var cookie = currHost.cookies[j];
              cookie.parentIndex = hostIndex;
              return cookie;
            }
            count++;
          }
        } else {
          // A host entry was open, but we weren't looking for an index
          // within that host entry's children, so skip forward over the
          // entry's children. We need to add one to increment for the
          // host value too.
          count += currHost.cookies.length + 1;
        }
      } else {
        count++;
      }

      for (let j = cacheStart; j < count; j++) {
        this._cacheItems[j] = cacheEntry;
      }
      this._cacheValid = count - 1;
    }
    return null;
  },

  _removeItemAtIndex: function (aIndex, aCount) {
    var removeCount = (typeof(aCount) == "undefined" ? 1 : aCount);
    if ((this._mode & VIEW_FLAT) == VIEW_FLAT) {
      // remove the cookies from the unfiltered set so that they
      // don't reappear when the filter is changed. See bug 410863.
      for (let i = aIndex; i < aIndex + removeCount; ++i) {
        var item = this._flatSet[i];
        var parent = gCookieKeeper._hosts[item.rawHost];
        for (let j = 0; j < parent.cookies.length; ++j) {
          if (item == parent.cookies[j]) {
            parent.cookies.splice(j, 1);
            break;
          }
        }
      }
      this._flatSet.splice(aIndex, removeCount);
    } else {
      var item = this._getItemAtIndex(aIndex);
      if (!item) return;
      this._invalidateCache(aIndex - 1);
      if (item.container) {
        delete gCookieKeeper._hosts[item.rawHost];

      } else {
        var parent = this._getItemAtIndex(item.parentIndex);
        for (let i = 0; i < parent.cookies.length; ++i) {
          var cookie = parent.cookies[i];
          if (item.rawHost == cookie.rawHost &&
              item.name == cookie.name && item.path == cookie.path)
            parent.cookies.splice(i, removeCount);
        }
      }
    }
  },

  _getIndexOfItem: function(aItem) {
    if (typeof(aItem) == "undefined" || aItem == null) {
      return -1;
    }

    if (this._mode == VIEW_TREE) {
      var hostOrder = gCookieKeeper._hostOrder;
      const nHosts = hostOrder.length;
      var index = 0;
      for (let cHost = 0; cHost < nHosts; cHost++) {
        var aHost = gCookieKeeper._hosts[hostOrder[cHost]];
        const isHost = aItem.rawHost == aHost.rawHost;
        if (aItem.container && isHost) {
          return index;
        }
        index++;
        if (isHost) {
          if (!aHost.open) {
            return -1;
          }
          const nCookies = aHost.cookies.length;
          for (let cCookie=0; cCookie<nCookies; cCookie++) {
            var cookie = aHost.cookies[cCookie];
            if (aItem.host == cookie.host && aItem.name == cookie.name && aItem.path == cookie.path) {
              return index;
            }
            index++;
          }
          // not found!
          return -1;
        } else if (aHost.open) {
          index += aHost.cookies.length;
        }
      }
    } else if (aItem.container) {
      return -1;
    } else {
      const nItems = this._flatSet.length;
      for (let cItem = 0; cItem < nItems; cItem++) {
        var cookie = this._flatSet[cItem];
        if (aItem.host == cookie.host && aItem.name == cookie.name && aItem.path == cookie.path) {
          return cItem;
        }
      }
    }
    return -1;
  },

  _invalidateCache: function (aIndex) {
    this._cacheValid = Math.min(this._cacheValid, aIndex);
  },

  _getCookieCellValue: function(aCookie, aColumnId) {
    if (aColumnId == COL_DOMAIN) {
      var aHost = aCookie.host
      return aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost
    } else if (aColumnId == COL_NAME) {
      return aCookie.name;
    } else if (aColumnId == COL_PROTECTED) {
      return aCookie.isProtected ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == COL_HTTPONLY) {
      return aCookie.isHttpOnly ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == COL_SECURE) {
      return aCookie.isSecure ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == COL_SESSION) {
      return aCookie.isSession ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == COL_VALUE) {
      return aCookie.value;
    } else if (aColumnId == COL_EXPIRE) {
      let expDate = aCookie.isSession ? aCookie.expiry : aCookie.expires;
      return expDate >= DATE_MAX ?
              gCookieKeeper._bundle.getString("date.max") :
              gCookieKeeper._formatDateString(expDate, "expireAtEndOfSession");
    } else if (aColumnId == COL_CREATION) {
      return gCookieKeeper._formatDateString(aCookie.creationTime, null);
    } else if (aColumnId == COL_LASTACCESS) {
      return gCookieKeeper._formatDateString(aCookie.lastAccessed, null);
    } else if (aColumnId == COL_PATH) {
      return aCookie.path;
    } else {
      return "n/a";
    }
  },

  getCellText: function (aIndex, aColumn) {
    if (this._mode == VIEW_TREE) {
      var item = this._getItemAtIndex(aIndex);

      if (typeof(item) == "undefined" || item == null) {
        return "";
      }

      if (item.container) {
        if (aColumn.id == COL_DOMAIN) {
          return item.rawHost;
        } else {
          return "";
        }
      } else {
        return this._getCookieCellValue(item, aColumn.id);
      }

    } else if (aIndex < this._flatSet.length) {
      var item = this._flatSet[aIndex];
      return this._getCookieCellValue(item, aColumn.id);
    } else {
      log.e("aIndex " + aIndex + ">= this._flatSet.length " + this._flatSet.length + " rowCount=" + this.rowCount);
    }
    return "";
  },

  get selection () { return this._selection; },
  set selection (val) {
    this._selection = val;
    return val;
  },

  getRowProperties: function (aIndex) {
    /*
     // FIXME set protected icon with CSS. Don't work for now, use getImageSrc() instead
    var item = this._getItemAtIndex(aIndex);
    if (item == null || item.container || !item.isProtected) {
      return "";
    }
    if (item.isProtected) {
      return "protected";
    }*/
    return "";
  },
  getCellProperties: function (aIndex, aColumn) { return ""; },
  getColumnProperties: function (aColumn) { return ""; },
  isContainer: function (aIndex) {
    if (this._mode == VIEW_TREE) {
      var item = this._getItemAtIndex(aIndex);
      if (!item) { return false; }
      return item.container;
    }
    return false;
  },
  isContainerOpen: function (aIndex) {
    if (this._mode == VIEW_TREE) {
      var item = this._getItemAtIndex(aIndex);
      if (!item) { return false; }
      return item.open;
    }
    return false;
  },
  isContainerEmpty: function (aIndex) {
    if (this._mode == VIEW_TREE) {
      var item = this._getItemAtIndex(aIndex);
      if (!item) return false;
      return item.cookies.length == 0;
    }
    return false;
  },
  isSeparator: function (aIndex) { return false; },
  isSorted: function (aIndex) { return false; },
  canDrop: function (aIndex, aOrientation) { return false; },
  drop: function (aIndex, aOrientation) {},
  getParentIndex: function (aIndex) {
    if (this._mode == VIEW_TREE) {
      var item = this._getItemAtIndex(aIndex);
      // If an item has no parent index (i.e. it is at the top level) this
      // function MUST return -1 otherwise we will go into an infinite loop.
      // Containers are always top level items in the cookies tree, so make
      // sure to return the appropriate value here.
      if (!item || item.container) return -1;
      return item.parentIndex;
    }
    return -1;
  },
  hasNextSibling: function (aParentIndex, aIndex) {
    if (this._mode == VIEW_TREE) {
      // |aParentIndex| appears to be bogus, but we can get the real
      // parent index by getting the entry for |aIndex| and reading the
      // parentIndex field.
      // The index of the last item in this host collection is the
      // index of the parent + the size of the host collection, and
      // aIndex has a next sibling if it is less than this value.
      var item = this._getItemAtIndex(aIndex);
      if (item) {
        if (item.container) {
          for (let i = aIndex + 1; i < this.rowCount; i++) {
            var subsequent = this._getItemAtIndex(i);
            if (subsequent.container) { return true; }
          }
          return false;
        } else {
          var parent = this._getItemAtIndex(item.parentIndex);
          if (parent && parent.container) {
            return aIndex < item.parentIndex + parent.cookies.length;
          }
        }
      }
    }
    return aIndex < this.rowCount - 1;
  },
  hasPreviousSibling: function (aIndex) {
    if (this._mode == VIEW_TREE) {

      var item = this._getItemAtIndex(aIndex);
      if (!item) { return false; }

      var parent = this._getItemAtIndex(item.parentIndex);
      if (parent && parent.container) {
        return aIndex > item.parentIndex + 1;
      }
    }
    return aIndex > 0;
  },
  getLevel: function (aIndex) {
    if (this._mode == VIEW_TREE) {
      var item = this._getItemAtIndex(aIndex);
      if (!item) { return 0; }
      return item.level;
    }
    return 0;
  },
  getImageSrc: function (aIndex, aColumn) {
    if (aColumn.index == 0) {
      var item = this._getItemAtIndex(aIndex);
      if (item && !item.container && item.isProtected) {
        return "chrome://cookiekeeper_os/skin/secure16.png";
      }
    }
    return "";
  },
  getProgressMode: function (aIndex, aColumn) {},
  getCellValue: function (aIndex, aColumn) {},
  setTree: function (aTree) {},
  toggleOpenState: function (aIndex) {
    if (this._mode == VIEW_TREE) {

      var item = this._getItemAtIndex(aIndex);
      if (item == null || !item.container) {
        return;
      }
      this._invalidateCache(aIndex);
      var multiplier = item.open ? -1 : 1;
      var delta = multiplier * item.cookies.length;
      this._rowCount += delta;
      item.open = !item.open;
      gCookieKeeper._tree.treeBoxObject.rowCountChanged(aIndex + 1, delta);
      gCookieKeeper._tree.treeBoxObject.invalidateRow(aIndex);
    }
  },
  cycleHeader: function (aColumn) {},
  selectionChanged: function () {},
  cycleCell: function (aIndex, aColumn) {},
  isEditable: function (aIndex, aColumn) {
    return false;
  },
  isSelectable: function (aIndex, aColumn) {
    return false;
  },
  setCellValue: function (aIndex, aColumn, aValue) {},
  setCellText: function (aIndex, aColumn, aValue) {},
  performAction: function (aAction) {},
  performActionOnRow: function (aAction, aIndex) {},
  performActionOnCell: function (aAction, aindex, aColumn) {}
}
