/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Import/Export cookies.
 *
 * ************************************************************************
 */

"use strict";

Cu.import("resource://gre/modules/NetUtil.jsm");
Cu.import("resource://gre/modules/FileUtils.jsm");

const DATE_MAX_SECONDS = 100000000 * 86400;
const TRUE        = "true";
const FALSE       = "false";
const EXPORT_ALL  = 1;
const EXPORT_PROT = 2;
const EXPORT_SEL  = 3;
const CK_CSV_HEADER = "host,path,name,value,creationTime,lastAccessed,expires,isDomain,isSecure,isHttpOnly,isSession,isProtected";

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/endsWith#Browser_compatibility
if (!String.prototype.endsWith) {
    Object.defineProperty(String.prototype, 'endsWith', {
        enumerable: false,
        configurable: false,
        writable: false,
        value: function (searchString, position) {
            position = position || this.length;
            position = position - searchString.length;
            var lastIndex = this.lastIndexOf(searchString);
            return lastIndex !== -1 && lastIndex === position;
        }
    });
}

var gImportExport = {

  _sCookieManager: Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2),
  _sPreferences  : Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch),

  onImport: function() {
    var bundle     = document.getElementById("bundleCookieKeeper");
    var filePicker = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);
    var fpCallback = function fpCallback_done(aResult) {
      if (aResult != Ci.nsIFilePicker.returnCancel && filePicker.fileURL) {
        gImportExport._import(filePicker.file.path);
      }
    }
    filePicker.appendFilter(bundle.getString("csv_file"), "*.csv");
    filePicker.init(window, bundle.getString("select_file"), Ci.nsIFilePicker.modeOpen);


    try {
      filePicker.open(fpCallback);
    } catch (ex) {
      let result = filePicker.show();
      fpCallback(result);
    }
  },

  onExportAll: function() {
    this._getExportFile(EXPORT_ALL);
  },

  onExportProtected: function() {
    this._getExportFile(EXPORT_PROT);
  },

  onExportSelected: function() {
    this._getExportFile(EXPORT_SEL);
  },

  _splitCSVLine: function(aLine, aLineNumber) {
    // Use a regex to parse the line?
    var idx    = 0;
    var datas  = [];
    var lineLen = aLine.length;

    let expectColSep = false;

    // log.d("parse line " + aLineNumber + " len="+lineLen+": " + aLine);

    while (idx < lineLen) {
      if (expectColSep || aLine[idx] == ',') {

        if (expectColSep && aLine[idx] != ',') {
          log.e("Line " + aLineNumber + ":" + (idx+1) + " expected comma, found \'" + aLine[idx] + "\'");
          idx = aLine.length;
          datas = [];
        } else if (!expectColSep) {
          // Empty column
          datas.push(null);
          // log.d("Line " + aLineNumber + ":" + (idx+1) + " empty column " + datas.length);
        }
        expectColSep = false;
        idx++;
      } else if (aLine[idx] == '\"') {
        // it is a quote protected column
        // search for the end of this protected string

        // log.d("Line " + aLineNumber + ":" + (idx+1) + ": found a quote protected string");

        let endIdx  = idx + 1;
        let nDQuote = 1;
        let isPcdq  = false; // Is previous char a double quote
        while (endIdx < lineLen && nDQuote > 0) {
          let ch = aLine[endIdx];
          log.d("Line " + aLineNumber + ":" + (endIdx+1) + ": ch=\'"+ch+"\'");
          if (ch == "\"") {
            nDQuote++;
            if (isPcdq) {
              nDQuote -= 2;
              isPcdq   = false;
            } else {
              isPcdq = true;
            }
          } else if (isPcdq) {
            // If it is a valid string normally we have reach the end here
            // The current char must be the comma.
            // log.d("Line " + aLineNumber + ":" + (endIdx+1) + ": isPcdq=true nDQuote="+nDQuote);
            nDQuote -= 2;
            isPcdq   = false;
          }
          endIdx++;
        }

        if ((nDQuote == 0) || (nDQuote == 2 && endIdx==lineLen)) { // && aLine[endIdx-1] == ','
          // Valid string
          endIdx = nDQuote == 0 ? endIdx-1 : endIdx;
          let data = aLine.substring(idx+1, endIdx-1);
          data = data.replace("\\r", "\r", "g");
          data = data.replace("\\n", "\n", "g");
          data = data.replace("\\\\", "\\", "g");
          data = data.replace("\"\"", "\"", "g");
          datas.push(data);
          idx = endIdx;
          expectColSep = true;
          // log.d("Line " + aLineNumber + ": endIdx=" + endIdx + " found quote protected string=\'" + data + "\'");
        } else {
          log.e("Line " + aLineNumber + ": Invalid quote string at pos " + (idx+1));
          // log.d("nDQuote="+nDQuote + " isPcdq="+isPcdq+" endIdx="+endIdx);
          idx = aLine.length;
          datas = [];
        }
      } else {
        // not a quote protected column
        let nextComma = aLine.indexOf(',', idx);
        let endIdx    = nextComma > -1 ? nextComma : aLine.length;
        let data      = aLine.substring(idx, endIdx);
        datas.push(data);
        // log.d("Line " + aLineNumber + ":" + (idx+1) + ": found normal column " + datas.length + " \'" + data + "\'");
        idx = endIdx;
        expectColSep = true;
      }
    }

    return datas;
  },

  _import: function(aPath) {
    log.i("import " + aPath);

    try {

      var importFile = null;
      try {
        importFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
        importFile.initWithPath(aPath);
      } catch (ex) {
        // Deprecated now, but needed for older version
        importFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
        importFile.initWithPath(aPath);
      }

      if (!importFile.isFile()) {
        return;
      }


      var istream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
      istream.init(importFile, -1, -1, false);

      var fileStr = NetUtil.readInputStreamToString(istream, istream.available(), {charset:"UTF-8", replacement:"?"});
      istream.close();
      log.d(fileStr);

      gCookieKeeper._blockCookieChangedEvent = true;
      gCookieKeeper._blockProtectionEvent = true;
      gCookieKeeper._view.selection.clearSelection();
      gCookieKeeper._saveState(); // will be restored in sort()

      var lines = fileStr.split("\n");
      let nImport = 0;
      let nTotal  = 0;
      let regexNumber = new RegExp("^[0-9]+$");
      let regexBool   = new RegExp("^(true|false)$");

      log.d("File have " + lines.length + " lines");

      for (let cLine=0; cLine<lines.length; cLine++) {
        let lineNumber = (cLine + 1).toString(10);
        let line = lines[cLine];

        if (line.length == 0 || (line.length > 1 && line[0] == '#') || CK_CSV_HEADER == line) {
          // skip this line
          log.i("Line " + lineNumber + ": empty, comment or header");
          continue;
        }
        nTotal++;
        let datas = this._splitCSVLine(line, lineNumber);
        if (datas.length != 12) {
          log.e("Line " + lineNumber + ": Invalid number of columns: expected 12 having " + datas.length);
          continue;
        }

        try {
          let path        = datas[1];
          let isProtected = (TRUE == datas[11]);

          if (path != null && path.length > 0) {
            // It is a cookie
            let host  = datas[0];
            let name  = datas[2];
            let value = datas[3];
            let sCreation = datas[4];
            let sAccessed = datas[5];
            let sDate     = datas[6]; // Expire date (string)
            let sIsDomain    = datas[7];
            let sIsSecure    = datas[8];
            let sIsHttpOnly  = datas[9];
            let sIsSession   = datas[10];
            let sIsProtected = datas[11];

            // Check datas:
            if (host == null || host.length == 0) {
              log.e("Line " + lineNumber + ": Invalid cookie host: empty");
              continue;
            }
            if (name == null || name.length == 0) {
              log.e("Line " + lineNumber + ": Invalid cookie name: empty");
              continue;
            }
            if (!regexNumber.test(sCreation)) {
              log.e("Line " + lineNumber + ": Invalid cookie creation date: not a number");
              continue;
            }
            if (!regexNumber.test(sAccessed)) {
              log.e("Line " + lineNumber + ": Invalid cookie last accessed date: not a number");
              continue;
            }
            if (!regexNumber.test(sDate)) {
              log.e("Line " + lineNumber + ": Invalid cookie expires date: not a number");
              continue;
            }
            if (!regexBool.test(sIsDomain)) {
              log.e("Line " + lineNumber + ": Invalid isDomain value: \'" + sIsDomain + "\', expected 'true' or 'false'");
              continue;
            }
            if (!regexBool.test(sIsSecure)) {
              log.e("Line " + lineNumber + ": Invalid isSecure value: \'" + sIsSecure + "\', expected 'true' or 'false'");
              continue;
            }
            if (!regexBool.test(sIsHttpOnly)) {
              log.e("Line " + lineNumber + ": Invalid isHttpOnly value: \'" + sIsHttpOnly + "\', expected 'true' or 'false'");
              continue;
            }
            if (!regexBool.test(sIsSession)) {
              log.e("Line " + lineNumber + ": Invalid isSession value: \'" + sIsSession + "\', expected 'true' or 'false'");
              continue;
            }
            if (!regexBool.test(sIsProtected)) {
              log.e("Line " + lineNumber + ": Invalid isProtected value: \'" + sIsProtected + "\', expected 'true' or 'false'");
              continue;
            }

            let isSecure   = (TRUE == sIsSecure);
            let isHttpOnly = (TRUE == sIsHttpOnly);
            let isSession  = (TRUE == sIsSession);
            // Convert expires date to a number
            let iDate = parseInt(sDate, 10);
            if (sDate.endsWith("000") && iDate != DATE_MAX_SECONDS) {
              // Workaround a previous bug #1297486
              iDate = (iDate/1000);
            } else if (iDate == 0 || isSession) {
              // Some session cookies have 0 for their expire date (previous export bug)
              iDate = DATE_MAX_SECONDS;
            }
            this._sCookieManager.add(host, path, name, value, isSecure, isHttpOnly, isSession, iDate);

            if (isProtected) {
              let cookie = {
                host      : host,
                name      : name,
                path      : path,
              };
              CookieKeeper.addProtectedCookie(cookie);
            }

            nImport++;
          } // else: probably a dom storage: import from CookieKeeper >= 1.8.0 ?
        } catch (ex) {
          log.e(ex);
        }
      }

      log.i(nImport.toString(10) + "/" + nTotal.toString(10) + " cookies imported");
    } catch (e) {
      var bundle        = document.getElementById("bundleCookieKeeper");
      var promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
      var msg           = "";
      var msgError      = "";
      var title         = bundle.getString("import_error");

      switch (e.result) {
      case Components.results.NS_ERROR_FILE_ACCESS_DENIED:
        msgError = bundle.getString("file_access_denied");
        break;
      case Components.results.NS_ERROR_FILE_IS_LOCKED:
        msgError = bundle.getString("file_is_locked");
        break;
      default:
        msgError = e.message;
        break;
      }

      msg = bundle.getFormattedString("read_file_failed", [aPath, msgError]);
      Cu.reportError(msg);
      Cu.reportError(e);
      promptService.alert(window, title, msg);
    }

    gCookieKeeper._blockCookieChangedEvent = false;
    gCookieKeeper._blockProtectionEvent = false;
    gCookieKeeper._refreshList(false);
    gCookieKeeper._tree.focus();
  },

  _getExportFile: function(aMethod) {
    var bundle     = document.getElementById("bundleCookieKeeper");
    var filePicker = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);
    var fpCallback = function fpCallback_done(aResult) {
      if (aResult != Ci.nsIFilePicker.returnCancel && filePicker.fileURL) {
        gImportExport._export(filePicker.file.path, aMethod);
      }
    }
    var date = new Date();
    var date_str =
      this.int2str(date.getFullYear(),    4) +
      this.int2str((date.getMonth() + 1), 2) +
      this.int2str(date.getDate(),        2) +
      "_"                                    +
      this.int2str(date.getHours(),       2) +
      this.int2str(date.getMinutes(),     2) +
      this.int2str(date.getSeconds(),     2);


    filePicker.appendFilter(bundle.getString("csv_file"), "*.csv");
    filePicker.defaultString = "cookies-" + date_str + ".csv";
    filePicker.defaultExtension = "csv";
    filePicker.init(window, bundle.getString("select_file"), Ci.nsIFilePicker.modeSave);
    try {
      filePicker.open(fpCallback);
    } catch (ex) {
      let result = filePicker.show();
      fpCallback(result);
    }
  },

  _getCookies: function(aMethod) {
    var cookies = [];
    if (aMethod == EXPORT_SEL) {
      var seln       = gCookieKeeper._view.selection;
      var rangeCount = seln.getRangeCount();
      for (let i=0; i<rangeCount; i++) {
        var min = {}; var max = {};
        seln.getRangeAt(i, min, max);
        for (let j = min.value; j <= max.value; j++) {
          var item = gCookieKeeper._view._getItemAtIndex(j);
          if (item.container) {
            var nCookies = item.cookies.length;
            for (let cCookies=0; cCookies<nCookies; cCookies++) {
              cookies.push(item.cookies[cCookies]);
            }
          } else {
            cookies.push(item);
          }
        }
      }
    } else {
      let ecookies = this._sCookieManager.enumerator;

      while (ecookies.hasMoreElements()) {
        let cookie = ecookies.getNext();
        if (cookie && cookie instanceof Ci.nsICookie2) {
          // Be sure to return the same object type than
          // when export the selection (see bug #1297486)
          let tc = gCookieKeeper._makeCookieObject(cookie.host, cookie, false);
          cookies.push(tc);
        }
      }
    }

    return cookies;
  },

  _getExportCookieLine: function(aCookie) {
    if (aCookie == null) {
      return null;
    }
    // deal with [,\n\r] chars in the string
    let value = aCookie.value;
    value = value.replace("\\", "\\\\", "g");
    value = value.replace("\n", "\\n",  "g");
    value = value.replace("\r", "\\r",  "g");
    value = value.replace("\"", "\"\"", "g");
    value = "\"" + value + "\"";

    // creationTime and lastAccessed are in microseconds since midnight (00:00:00), January 1, 1970 UTC.
    // expires and expiry are in seconds since midnight (00:00:00), January 1, 1970 UTC...
    // Export everything in seconds.
    let creationDate = Math.floor(aCookie.creationTime / 1000).toString(10);
    let accessedDate = Math.floor(aCookie.lastAccessed / 1000).toString(10);
    let expires      = (aCookie.isSession ? aCookie.expiry : aCookie.expires);
    let expiresDate  = Math.floor(expires/1000).toString(10);

    let str = aCookie.host   +
          "," + aCookie.path +
          "," + aCookie.name +
          "," + value        +
          "," + creationDate +
          "," + accessedDate +
          "," + expiresDate  +
          "," + (aCookie.isDomain    ? TRUE : FALSE) +
          "," + (aCookie.isSecure    ? TRUE : FALSE) +
          "," + (aCookie.isHttpOnly  ? TRUE : FALSE) +
          "," + (aCookie.isSession   ? TRUE : FALSE) +
          "," + (aCookie.isProtected ? TRUE : FALSE) +
          "\n";
    return str;
  },

  _export: function(aPath, aMethod) {

    log.i("Export to " + aPath);

    try {

      var exportFile = null;
      try {
        exportFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
        exportFile.initWithPath(aPath);
      } catch (ex) {
        exportFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
        exportFile.initWithPath(aPath);
      }

      if (exportFile.exists()) {
        exportFile.remove(false);
      }
      exportFile.create(Ci.nsIFile.NORMAL_FILE_TYPE, parseInt("600", 8)); // Unix file perm: -rw-------

      let ostream = FileUtils.openFileOutputStream(exportFile, FileUtils.MODE_WRONLY | FileUtils.MODE_CREATE);

      let converter = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Components.interfaces.nsIConverterOutputStream);
      converter.init(ostream, "UTF-8", 0, 0);

      let pcookies = CookieKeeper.loadProtectedCookies();
      let cookies  = this._getCookies(aMethod);

      let str = CK_CSV_HEADER + "\n";
      converter.writeString(str, str.length);

      let nCookies  = cookies.length;
      let nPCookies = pcookies.length; // number of protected cookie
      let lPCookie  = nPCookies - 1;   // last protected cookie (use to move already matched pcookies to the end of the list)
      let nExport   = 0; // number of exported cookies

      for (let cCookies = 0; cCookies<nCookies; cCookies++) {
        let cookie = cookies[cCookies];
        let isProtected = false;
        for (let cProt=0; cProt<nPCookies && lPCookie>-1; cProt++) {
          let pcookie = pcookies[cProt];
          if (pcookie.host == cookie.host && pcookie.name == cookie.name && pcookie.path == cookie.path) {
            isProtected = true;
            // move this pcookie to the end of the list
            pcookies[cProt]    = pcookies[lPCookie];
            pcookies[lPCookie] = pcookie;
            lPCookie--;
            cProt = nPCookies;
            cookie.isProtected = true;
          }
        }

        if (!(aMethod == EXPORT_PROT && !isProtected)) {
          str = this._getExportCookieLine(cookie);
          converter.writeString(str, str.length);
          nExport++;
        }
      }
      converter.close();
      gCookieKeeper._tree.focus();
      log.i(nExport + " cookies exported");
    } catch (e) {
      var bundle        = document.getElementById("bundleCookieKeeper");
      var promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
      var msgError      = "";
      var msg           = "";
      var title         = bundle.getString("export_error");

      switch (e.result) {
      case Components.results.NS_ERROR_FILE_ACCESS_DENIED:
        msgError = bundle.getString("file_access_denied");
        break;
      case Components.results.NS_ERROR_FILE_IS_LOCKED:
        msgError = bundle.getString("file_is_locked");
        break;
      case Components.results.NS_ERROR_FILE_NO_DEVICE_SPACE:
      case Components.results.NS_ERROR_FILE_DISK_FULL:
        msgError = bundle.getString("write_file_no_device_space");
        break;
      default:
        msgError = e.message;
        break;
      }

      msg = bundle.getFormattedString("write_file_failed", [aPath, msgError]);
      Cu.reportError(msg);
      Cu.reportError(e);
      promptService.alert(window, title, msg);
    }
  },

  int2str: function(aNumber, aPrecision) {
    var s = "" + aNumber;
    while (s.length < aPrecision) {
      s = "0" + s;
    }

    return s;
  }
}
