/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Public API
 *
 * ************************************************************************
 */

"use strict";

var EXPORTED_SYMBOLS = ["CookieKeeper"];

let {Services} = Components.utils.import("resource://gre/modules/Services.jsm");

try {
  Components.utils.import("resource://gre/modules/devtools/Console.jsm");
} catch (ex) {
  // Silently ignore
}

let self = this;
Services.scriptloader.loadSubScript("chrome://cookiekeeper/content/common.js", self);

let sPrefs    = Services.prefs;
let sObserver = Services.obs;

/**
 * @brief Common/Public CookieKeeper function.
 *
 * @par Observer example:
 * see https://developer.mozilla.org/en-US/Add-ons/Overlay_Extensions/XUL_School/Observer_Notifications
 ~~~~~~~~~~~~~{.js}
    let pCObserver = {
      observe : function(aSubject, aTopic, aData) {
          if (aTopic == CookieKeeper.OBS_TOPIC) {
            if (aData == CookieKeeper.OBS_STARTUP || aData == CookieKeeper.OBS_SHUTDOWN) {

            } else {
              var pCookie = aSubject.wrappedJSObject.pcookie;
              var action  = "added";
              if (aData == CookieKeeper.OBS_DEL) {
                action = "removed";
              }
              window.alert("PCookie " + action + ": " + pCookie.host + ":" + pCookie.name);
          }
        }
      }
 ~~~~~~~~~~~~~
 */

var CookieKeeperPrivate = {
  getCookieWindows: function() {
    var e    = Services.wm.getEnumerator(null);
    var wins = [];

    while (e.hasMoreElements()) {
      let win = e.getNext().QueryInterface(Components.interfaces.nsIDOMChromeWindow);
      let id  = win.document.documentElement.getAttribute("id");

      for (let i=0; i<CK_WINDOWS_ID.length; i++) {
        if (CK_WINDOWS_ID[i] == id) {
          wins.push({id: id, window:win});
        }
      }
    }

    return wins;
  }
}

var CookieKeeper = {
  API_VERSION:  1,                        /// The CookieKeeper public API version
  OBS_TOPIC:    CK_OBS_PCOOKIE_TOPIC,     /// Observer ProtectedCookies topic
  OBS_STARTUP:  CK_OBS_PCOOKIE_STARTUP,   /// Add-on is starting
  OBS_SHUTDOWN: CK_OBS_PCOOKIE_SHUTDOWN,  /// Add-on is shutting down
  OBS_ADD:      CK_OBS_PCOOKIE_ADD,       /// A protected cookie is added
  OBS_DEL:      CK_OBS_PCOOKIE_DEL,       /// A protected cookie is removed

  // Preferences panel open options:
  PREFPANE_OPTIONS  : CK_PREFPANE_OPTIONS,      /// To show the CookieKeeper preference pane
  PREFPANE_PCOOKIES : CK_PREFPANE_PCOOKIES,     /// To show the protected cookies pane
  PREFPANE_PERMS    : CK_PREFPANE_PERMISSIONS,  /// To show the cookies permissions pane
  PREFPANE_BROWSER  : CK_PREFPANE_BROWSER,      /// To show the browser preferences pane

  /**
   * @brief Called only by the CookieKeeper bootstrap startup() method
   */
  init: function() {
    sObserver.notifyObservers(null, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_STARTUP);
    log.d("CookieKeeper resource: init done");
  },

  /**
   * @brief Called only by the CookieKeeper bootstrap shutdown() method
   */
  uninit: function() {
    sObserver.notifyObservers(null, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_SHUTDOWN);
    log.d("CookieKeeper resource: uninit done");
  },

  /**
   * @brief Show the main CookieKeeper window. If the dialogue is already opened, it will be focus.
   * @param window The DOM window.
   */
  showMainWindow: function(window) {
    var ckws = CookieKeeperPrivate.getCookieWindows();
    var opened = false;
    for (let i=0; i<ckws.length && opened == false; i++) {
      let ckw = ckws[i];
      if (ckw.id == CK_WINDOWS_ID[CK_WINDOW_MAIN_IDX]) {
        ckw.window.focus();
        opened = true;
      }
    }

    if (!opened) {
      window.open(
          "chrome://cookiekeeper/content/cookiekeeper.xul",
          "cookieKeeperWindow",
          "chrome,toolbar,centerscreen,resizable"
      );
    }
  },

  /**
   * @brief Show the CookieKeeper preferences dialogue. If the window is already opened, it will be focus.
   * @param window The DOM window.
   * @param aPane The pane to show (PREFPANE_OPTIONS/PREFPANE_PCOOKIES/PREFPANE_PERMS/PREFPANE_BROWSER)
   */
  showPreferencesDialogue: function(window, aPane) {
    var ckws = CookieKeeperPrivate.getCookieWindows();
    var opened = false;
    for (let i=0; i<ckws.length && opened == false; i++) {
      let ckw = ckws[i];
      if (ckw.id == CK_WINDOWS_ID[CK_WINDOW_PREFS_IDX]) {
        ckw.window.focus();
        opened = true;
      }
    }

    if (!opened) {
        let xulRuntime = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULRuntime);
        let osOptions = "";
        if (xulRuntime.OS == "WINNT") {
          // MS Windows love modal dialogue
          osOptions = ",modal";
        } else if (xulRuntime.OS == "Darwin") {
          // Don't really know why, but recommanded by Mozilla on MacOSX
          osOptions = ",dialog=yes";
        } // else {} // GNU/Linux don't like restrictions ;-)
        window.openDialog("chrome://cookiekeeper/content/preferences.xul",
                          "cookieKeeperPrefsWindow",
                          "chrome,titlebar,toolbar,centerscreen" + osOptions,
                          aPane == null ? CK_PREFPANE_OPTIONS : aPane);
    }
  },

  /**
   * @brief Close all CookieKeeper windows.
   */
  closeCookieWindow: function() {
    function timeoutClose(aWin) {
      try {
        aWin.close();
        if (!aWin.closed) {
          aWin.setTimeout(function() { timeoutClose(aWin); }, 50);
        }
      } catch (ex) {
        Components.utils.reportError(ex);
      }
    }

    var ckws = CookieKeeperPrivate.getCookieWindows();
    for (let i=ckws.length-1; i>=0; i--) {
      let ckw = ckws[i];
      timeoutClose(ckw.window);
    }
  },

  /**
   * @brief Load protected cookies list from the CookieKeeper preferences
   * @param psvc The preference service
   * @return an array of protected cookies.
   */
  loadProtectedCookies: function() {
    var protCookies = [];
    log.d("loadProtectedCookies()");
    if (sPrefs.prefHasUserValue(PREF_PROT_COOKIES)) {
      let sList = sPrefs.getCharPref(PREF_PROT_COOKIES);
      let cList = sList.split(PROT_SEP_COOKIE);
      for (let i=0; i<cList.length; i++) {
        if (cList[i].length > 0) {
          let cData = cList[i].split(PROT_SEP_DATA);
          let c = { host : cData[0],
                    name : cData[1],
                    path : cData[2]
                  };
          protCookies.push(c);
        }
      }
    }
    return protCookies;
  },

  /**
   * @brief Test if a cookie is protected
   * @param aCookie a cookie (having host/name/path attributes)
   * @param protList a list of protected cookies returned by loadProtectedCookies(), or null
   * @return true if the cookie is protected
   */
  isProtectedCookie: function(aCookie, protList) {
    var protCookies = null;
    if (protList == null) {
      protCookies = this.loadProtectedCookies();
    } else {
      protCookies = protList;
    }
    var isProtected = false;

    for (let i=0; i<protCookies.length && !isProtected; i++) {
      var pc = protCookies[i];
      isProtected = pc.host == aCookie.host && pc.name == aCookie.name && pc.path == aCookie.path;
    }

    return isProtected;
  },

  /**
   * @brief Add a protected cookie to the list
   * @param aCookie A protected cookie (host/name/path)
   * @return true if the protected cookie has been added
   * @par
   * Notify the observer: OBS_TOPIC/OBS_ADD
   */
  addProtectedCookie: function(aCookie) {
    if (aCookie == null || aCookie.container) {
      log.e("addProtectedCookie: Not a cookie");
      return false;
    }

    // aCookie.isProtected = true;

    const cString = aCookie.host + PROT_SEP_DATA + aCookie.name + PROT_SEP_DATA + aCookie.path + PROT_SEP_COOKIE;
    const invalidString = ";undefined;undefined"; // TODO must investigate when it happend
    var sList = PROT_SEP_COOKIE;

    if (cString == invalidString) {
      log.e("can't add invalid cookie: " + invalidString);
      return false;
    }

    if (sPrefs.prefHasUserValue(PREF_PROT_COOKIES)) {
      sList = sPrefs.getCharPref(PREF_PROT_COOKIES);
    }

    let idx = sList.indexOf(cString);
    if (idx == -1 || sList.charAt(idx - 1) != PROT_SEP_COOKIE) {
      // Only add cookie if it is not present
      log.i("add protected cookie: " + cString);
      sList = sList + cString;

      sPrefs.setCharPref(PREF_PROT_COOKIES, sList);

      // Notify observers
      var pCookie = {
        host: aCookie.host,
        name: aCookie.name,
        path: aCookie.path,
      }
      var data = {
        pcookie : pCookie
      }
      data.wrappedJSObject = data;
      sObserver.notifyObservers(data, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_ADD);

      return true;
    }

    log.e("protected cookie already added");
    return false;
  },

  /**
   * @brief Remove a protected cookie
   * @param aCookie A protected cookie (host/name/path)
   * @return true if the protected cookie has been removed
   * @par
   * Notify the observer: OBS_TOPIC/OBS_DEL
   */
  removeProtectedCookie: function(aCookie) {
    var protCookies  = this.loadProtectedCookies();
    var sList        = PROT_SEP_COOKIE;
    var pCookieFound = false;

    if (aCookie == null || aCookie.container) {
      log.e("remove protected cookie: not a cookie")
      return false;
    }

    for (let i=0; i<protCookies.length; i++) {
      let c = protCookies[i];

      if (!(c.host == aCookie.host && c.name == aCookie.name && c.path == aCookie.path)) {
        sList = sList + c.host + PROT_SEP_DATA + c.name + PROT_SEP_DATA + c.path + PROT_SEP_COOKIE;
      } else {
        pCookieFound = true;
        log.i("remove protected cookie: " + aCookie.host + " / " + aCookie.name + " / " + aCookie.path);
      }
    }

    if (pCookieFound) {
      // Update the prefs
      sPrefs.setCharPref(PREF_PROT_COOKIES, sList);

      // Notify observers
      var pCookie = {
        host: aCookie.host,
        name: aCookie.name,
        path: aCookie.path,
      }
      var data = {
        pcookie : pCookie
      }
      data.wrappedJSObject = data;
      sObserver.notifyObservers(data, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DEL);
    }
    return pCookieFound;
  }

}
