/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Preferences managment
 *
 * ************************************************************************
 */

"use strict";

let ckPreferences = {
  // Initializes default preferences
  // https://github.com/mozilla/addon-sdk/blob/85511f1c467a9fb5d2b8d2b6b4360e7598fce638/python-lib/cuddlefish/app-extension/bootstrap.js
  setDefaultPrefs: function() {
    const prefsDefault  = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).QueryInterface(Ci.nsIPrefBranch);

    let branchD = prefsDefault.getDefaultBranch("");
    let branchS = Services.prefs.getDefaultBranch("");
    let prefLoaderScope = {
      pref: function(key, val) {
        // set the value only if it does not exists
        var exists = branchS.prefHasUserValue(key);
        switch (typeof val) {
          case "boolean":
            branchD.setBoolPref(key, val);
            if (!exists) branchS.setBoolPref(key, val);
            break;
          case "number":
            branchD.setIntPref(key, val);
            if (!exists) branchS.setIntPref(key, val);
            break;
          case "string":
            branchD.setCharPref(key, val);
            if (!exists) branchS.setCharPref(key, val);
            break;
        }
      }
    };

    let prefjs = addonData.resourceURI.spec + "defaults/preferences/preferences.js";
    // setup default prefs
    Services.scriptloader.loadSubScript(prefjs, prefLoaderScope);
  },

  // Migrate CookieCuller preferences
  migrateCCPrefs: function() {
    function makePCookie(aArray) {
      var c = null;
      if (aArray != null && aArray.length == 3) {
        c = { host   : aArray[0],
              name   : aArray[1],
              path   : aArray[2],
              toStr  : function() { return this.host + PROT_SEP_DATA + this.name + PROT_SEP_DATA + this.path; }
        };
      }
      return c;
    }

    var psvc = Services.prefs;

    // Migrate & delete old cookie culler (v1.4) preferences
    const PREF_BRANCH_COOKIE_CULLER = "extensions.cookieculler.";
    const PREF_DEL_AT_STARTUP       = "extensions.cookieculler_options.delUnprotAtStartup";

    // Get protected cookies
    var bsvc   = Cc["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
    var branch = bsvc.getBranch(PREF_BRANCH_COOKIE_CULLER);

    if (branch.prefHasUserValue("count") || psvc.prefHasUserValue(PREF_DEL_AT_STARTUP)) {
      // Init the locale bundle
      l10n.init(addonData.resourceURI.spec, "cookiekeeper.properties");

      var promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
      var doMigr = promptService.confirm(null, "CookieKeeper", tr("install.prefs.cookieculler"));

      if (doMigr && branch.prefHasUserValue("count")) {
        let count       = branch.getIntPref("count");
        let protCookies = [];
        let pcListStr   = PROT_SEP_COOKIE;

        // Get our protected cookies
        if (psvc.prefHasUserValue(PREF_PROT_COOKIES)) {
          let sList = psvc.getCharPref(PREF_PROT_COOKIES);
          let cList = sList.split(PROT_SEP_COOKIE);
          for (let i=0; i<cList.length; i++) {
            if (cList[i].length > 0) {
              let cData = cList[i].split(PROT_SEP_DATA);
              protCookies.push(makePCookie(cData));
            }
          }
        }

        // Get CookieCuller protected cookies
        for (let i=0; i<count; i++) {
          const pname = "pcookie" + i;
          if (branch.prefHasUserValue(pname)) {
            let pstr = branch.getCharPref(pname);
            let cData = pstr.split(";");
            protCookies.push(makePCookie(cData));

          }
        }

        // Build the protection list
        for (let i=0; i<protCookies.length; i++) {
          let pcookie = protCookies[i];
          if (pcookie != null) {
            let pstr = protCookies[i].toStr() + PROT_SEP_COOKIE;
            if (pcListStr.indexOf(PROT_SEP_COOKIE + pstr) == -1) {
              log.d("Add protected cookie " + pstr);
              pcListStr = pcListStr + pstr;
            }
          }
        }

        log.i("install: migrate old cookies: " + pcListStr);
        psvc.setCharPref(PREF_PROT_COOKIES, pcListStr);
        psvc.deleteBranch(PREF_BRANCH_COOKIE_CULLER);
      } // do pcookies migration

      var delAtStartup = false;
      if (doMigr && psvc.prefHasUserValue(PREF_DEL_AT_STARTUP)) {
        delAtStartup = psvc.getBoolPref(PREF_DEL_AT_STARTUP);
        psvc.setBoolPref(PREF_CLEAR_COOKIES_AT_EXIT, delAtStartup);
        psvc.deleteBranch("extensions.cookieculler_options.");
      }
    }
  },

  updatePreferences: function() {
    var psvc     = Services.prefs;
    var oldVer   = CK_PREF_VERSION;
    var havePref = false;

    if (psvc.prefHasUserValue(PREF_VERSION)) {
      havePref = true;
      oldVer   = psvc.getIntPref(PREF_VERSION);
      if (oldVer == 150) {
        oldVer = 1;
      }
    }

    if (psvc.prefHasUserValue(PREF_PROT_COOKIES) && !havePref) {
      oldVer = 1; // bug lp:1305252
    }

    switch (oldVer) {
    case 1: {
      // Add the PROT_SEP_COOKIE char "|" at the end of the cookies list.
      // See bug lp:1282328 https://bugs.launchpad.net/cookiekeeper/+bug/1282328
      if (psvc.prefHasUserValue(PREF_PROT_COOKIES)) {
        let sList = psvc.getCharPref(PREF_PROT_COOKIES);
        if (sList.length > 1 && sList.charAt(sList.length - 1) != PROT_SEP_COOKIE) {
          sList = sList + PROT_SEP_COOKIE;
          psvc.setCharPref(PREF_PROT_COOKIES, sList);
        }
      }

      // fall-through
    }
    case 2:
      // fall-through
    default:
      break;
    } // switch

    // Update version
    if (oldVer != CK_PREF_VERSION || !havePref) {
      psvc.setIntPref(PREF_VERSION, CK_PREF_VERSION);
    }
  } // updatePreferences
} // ckPreferences
